from fastapi import UploadFile, File, Form
from typing import Optional, Union
from pathlib import Path
from git import Repo

from langchain.document_loaders import CSVLoader
from langchain.document_loaders import Docx2txtLoader
from langchain.document_loaders import PyMuPDFLoader
from langchain.document_loaders import TextLoader
from langchain.document_loaders import GitLoader
from langchain.document_loaders import WebBaseLoader

from langchain.text_splitter import RecursiveCharacterTextSplitter

import json
import os
import shutil

from .document_model import Document, SplitDocumentInput, UrlDocumentInput
from ...common_models import (
    UserProperties,
    RedisKeyType,
    DEFAULT_USER_ID,
    DEFAULT_SESSION_ID,
    DEFAULT_GIT_FOLDER,
)
from ..base_node import BaseNode, NodeConfig
from ...service.redis import Redis
from ....utils.logger import get_logger


logger = get_logger(__name__)

document_loader_config = {
    "name": "document_loader",
    "description": "A node that is capable of reading various document types. Redis key format would be user_id:session_id:documents",
    "functions": {
        "create_document_from_file": "Create a document from a given file.",
        "create_document_from_url": "Create a document from a given URL.",
        "split_documents": "Split the documents from the file with default chunk size and overlap.",
        "process_document": "Process the file's content into text",
        "remove_document": "Remove the document from the Redis and local.",
    },
}


class DocumentLoaderNode(BaseNode):
    config: NodeConfig = NodeConfig(**document_loader_config)

    def __init__(self):
        super().__init__()
        self.redis = Redis()

    def create_document_from_file(
        self,
        input: UploadFile = File(...),
        properties: UserProperties = Form(...),
    ):
        document = Document.create_document_from_file(input, properties)
        return self.process_document(document)

    def create_document_from_url(
        self,
        input: UrlDocumentInput,
        properties: UserProperties = Form(...),
    ):
        document = Document.create_document_from_url(input, properties)
        return self.process_document(document)

    def split_documents(self, input: SplitDocumentInput):
        if input.file_id is None and input.document is None:
            logger.error("Either file_id or document must be provided.")
            return None
        elif input.file_id is not None:
            logger.debug(f"Spliting documents for: {input.file_id}")
            for doc_json in self.redis.safe_get_with_key_type(
                input.user_properties, RedisKeyType.DOCUMENTS
            ):
                doc = Document(**json.loads(doc_json))
                if doc.file_id == input.file_id:
                    return self._split_documents(
                        doc, input.chunk_size, input.chunk_overlap
                    )
            return None
        elif input.document is not None:
            return self._split_documents(
                input.document, input.chunk_size, input.chunk_overlap
            )
        else:
            raise ValueError("Unknown error")

    def process_document(self, input: Document):
        """
        Process the file to compute the documents

        Args:
            input (Document): The document to process
            save_redis (bool, optional): Whether to save the document to Redis. Defaults to True.
        """
        file_path = str(input.file_path)
        if input.file_extension == "csv":
            loader = CSVLoader(file_path)
        elif input.file_extension == "docx":
            loader = Docx2txtLoader(file_path)
        elif input.file_extension == "pdf":
            loader = PyMuPDFLoader(file_path)
        elif input.file_extension == "txt":
            loader = TextLoader(file_path)
        elif input.file_extension == "web":
            loader = WebBaseLoader(input.file_name)
        elif input.file_extension == "git":
            repo_path = input.file_path
            repo_path_str = str(repo_path) + "/"
            repo_path.parent.mkdir(parents=True, exist_ok=True)
            repo = Repo.clone_from(input.file_name, repo_path_str)
            branch = repo.head.reference
            gitignore_file = os.path.join(repo_path_str, ".gitignore")
            if os.path.exists(gitignore_file):
                os.remove(gitignore_file)
            with open(gitignore_file, "w") as fp:
                pass

            loader = GitLoader(repo_path=repo_path_str, branch=branch)
            # TODO: not sure why some repo gitignore would ignore all files
            # loader = GitLoader(
            #     clone_url=input.file_name,
            #     repo_path=str(repo_path) + "/",
            #     branch="main",
            #     file_filter=lambda file_path: file_path.endswith(".py")
            # )
        elif input.file_extension == "json":
            # TODO: JSONLoader needs to input json schema metadata
            # loader = TextLoader(
            #     file_path=input.file_path,
            # )
            pass
        else:
            raise ValueError(f"Unsupported file extension: {input.file_extension}")

        documents = loader.load()
        input.documents = documents
        if not self.redis.exists_with_key_type(
            input.user_properties, RedisKeyType.DOCUMENTS
        ):
            logger.debug(
                f"Creating new knowledge base for {input.user_properties.user_id}:{input.user_properties.session_id} in Redis."
            )
            self.redis.safe_set_with_key_type(
                input.user_properties, RedisKeyType.DOCUMENTS, []
            )

        new_documents = self.redis.safe_get_with_key_type(
            input.user_properties, RedisKeyType.DOCUMENTS
        )
        new_documents.append(input.json())

        self.redis.safe_set_with_key_type(
            input.user_properties, RedisKeyType.DOCUMENTS, new_documents
        )
        return input.json()

    def remove_document(self, input: Document):
        if input.file_path is not None:
            if os.path.isfile(input.file_path):
                os.remove(input.file_path)
            elif os.path.isdir(input.file_path):
                shutil.rmtree(input.file_path)  # remove dir and all contains

        type = RedisKeyType.DOCUMENTS
        if not self.redis.exists_with_key_type(input.user_properties, type):
            logger.error(f"Document {input.file_name} does not exist.")
            return

        all_documents = self.redis.safe_get_with_key_type(input.user_properties, type)
        processed_documents = []
        for doc_json in all_documents:
            doc = Document(**json.loads(doc_json))
            if doc.file_id != input.file_id:
                processed_documents.append(doc_json)

        self.redis.safe_set_with_key_type(
            input.user_properties, type, processed_documents
        )

    def _split_documents(
        self,
        input: Document,
        chunk_size: Optional[int] = 500,
        chunk_overlap: Optional[int] = 0,
    ):
        """
        Split the documents from the file
        """
        logger.debug(f"Spliting documents from file {input.file_name}")

        text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )

        documents = text_splitter.split_documents(input.documents)
        return documents



import git
from ..base_node import BaseNode


class GitBaseNode(BaseNode):
    def __init__(self):
        # if len(args) > 0:
        #     self.repo_path = args[0]
        # else:
        #     if 'path' not in kwargs:
        #         raise ValueError("Missing argument: path to git repo at GitBaseNode().")
        #     self.repo_path = kwargs.get('path', '')
        # if type(self.repo_path) is not str:
        #     raise ValueError("Type error: argument path must be str.")

        self.g = git
        self.r = git.Repo
        super().__init__()



EXTRACT_PARAMETES_PROMPT="""
As a text parameter matcher, your job is to install the function parameter requirements from the text to extract the corresponding parameters.
1. You need to read and understand the following text(input_text):
2. Extract the corresponding content from the text according to the following parameter information:
3. Follow these steps:

- First, read and understand the text carefully.
   - Second, look for the appropriate content in the text based on the parameter information provided.
   - Next, generate a parameter reference dictionary based on what you find. If you don't find a match for a parameter in the text, leave it blank in the dictionary.

Perform the preceding steps in sequence and generate a dictionary of parameter parameters.

Textual content (input_text):
[insert text content here]

Parameter information (parameter_info):
[insert parameter info here]

Generated parameter reference dictionary:
[Generate a parameter reference dictionary here, and if the corresponding value cannot be matched, an empty dictionary will be generated]
"""
EXTRACT_PARAMETES_EXAMPLE_PROMPT="""
Here is some examples

Textual content (input_text):
"Alice, who is 25 years old, works at Acme Corp. Her email is alice@example.com."

Parameter information (parameter_info):
[{'name': 'name', 'required': True, 'parameter_schema': {'type': 'string', 'description': 'The name of this people.', 'properties': None, 'items': None}},{'name': 'age', 'required': True, 'parameter_schema': {'type': 'int', 'description': 'The age of this people.', 'properties': None, 'items': None}},{'name': 'email', 'required': True, 'parameter_schema': {'type': 'string', 'description': 'The email of this people.', 'properties': None, 'items': None}}]

Generated parameter reference dictionary:
{"name": "Alice", "age": 25, "email": "alice@example.com"}

Textual content (input_text):
"The weather in Paris is mostly sunny this week, with a high of 18°C and a low of 7°C."

Parameter information (parameter_info):
[{'name': 'w', 'required': True, 'parameter_schema': {'type': 'string', 'description': 'What kind of weather?', 'properties': None, 'items': None}},{'name': 'age', 'required': True, 'parameter_schema': {'type': 'int', 'description': 'The age of this people.', 'properties': None, 'items': None}},{'name': 'email', 'required': True, 'parameter_schema': {'type': 'string', 'description': 'The email of this people.', 'properties': None, 'items': None}}]

Generated parameter reference dictionary:
{"w": "sunny"}

Textual content (input_text):
"This new novel is thrilling and full of unexpected twists."

Parameter information (parameter_info):
[{'name': 'x', 'required': True, 'parameter_schema': {'type': 'string', 'description': 'input x content', 'properties': None, 'items': None}},{'name': 'age', 'required': True, 'parameter_schema': {'type': 'int', 'description': 'The age of this people.', 'properties': None, 'items': None}},{'name': 'email', 'required': True, 'parameter_schema': {'type': 'string', 'description': 'The email of this people.', 'properties': None, 'items': None}}]

Generated parameter reference dictionary:
{}
"""
EXTRACT_PARAMETES_HINT="""
Please provide the information in dictionary format. Make sure that the name of each parameter appears as a key, and extract its matching information from the text as the corresponding value.
"""


from ..base_node import BaseNode, NodeConfig
from .github_node import GithubNode
from .github_model import (
    GetIssueInput,
    CreateIssueCommentInput,
    CreateIssueInput,
    CreateIssueWithBodyInput,
    CreateIssueWithLabelsInput,
    CreateIssueWithAssigneeInput,
    CreateIssueWithMilestoneInput,
    CloseAllIssuesInput,
)

from github import Github

github_issues_node_config = {
    "name": "github_issues",
    "description": "A node for interacting with GitHub issues.",
    "functions": {
        "get_issue": "Get issue details.",
        "create_issue_comment": "Create a comment on an issue.",
        "create_issue": "Create a new issue.",
        "create_issue_with_body": "Create a new issue with a body.",
        "create_issue_with_labels": "Create a new issue with labels.",
        "create_issue_with_assignee": "Create a new issue with an assignee.",
        "create_issue_with_milestone": "Create a new issue with a milestone.",
        "close_all_issues": "Close all open issues in a repository.",
    },
}


class GithubIssuesNode(GithubNode):
    config: NodeConfig = NodeConfig(**github_issues_node_config)

    def __init__(self):
        super().__init__()  # This will call the GithubNode's __init__ method and set up the token and GitHub instance

    def get_issue(self, input: GetIssueInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            issue = repo.get_issue(number=input.issue_number)
            return issue.title, issue.body
        except Exception as e:
            return str(e)

    def create_issue_comment(self, input: CreateIssueCommentInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            issue = repo.get_issue(number=input.issue_number)
            comment = issue.create_comment(input.comment_content)
            return f"Comment created with ID: {comment.id}"
        except Exception as e:
            return str(e)

    def create_issue(self, input: CreateIssueInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            issue = repo.create_issue(title=input.title)
            return f"Issue created with title: {issue.title} and ID: {issue.id}"
        except Exception as e:
            return str(e)

    def create_issue_with_body(self, input: CreateIssueWithBodyInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            issue = repo.create_issue(title=input.title, body=input.body)
            return f"Issue created with title: {issue.title}, ID: {issue.id}, and body: {issue.body}"
        except Exception as e:
            return str(e)

    def create_issue_with_labels(self, input: CreateIssueWithLabelsInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            labels = [repo.get_label(label_name) for label_name in input.labels]
            issue = repo.create_issue(title=input.title, labels=labels)
            return f"Issue created with title: {issue.title}, ID: {issue.id}, and labels: {[label.name for label in issue.labels]}"
        except Exception as e:
            return str(e)

    def create_issue_with_assignee(self, input: CreateIssueWithAssigneeInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            issue = repo.create_issue(title=input.title, assignee=input.assignee)
            return f"Issue '{issue.title}' created with assignee '{input.assignee}' successfully."
        except Exception as e:
            return str(e)

    def create_issue_with_milestone(self, input: CreateIssueWithMilestoneInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            # Retrieve the milestone by its number
            milestone = repo.get_milestone(number=input.milestone_number)
            # Create an issue with the milestone
            issue = repo.create_issue(
                title=input.title, body=input.body, milestone=milestone
            )
            return {
                "status": "Issue created successfully with milestone.",
                "issue_number": issue.number,
            }
        except Exception as e:
            return str(e)

    def close_all_issues(self, input: CloseAllIssuesInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            open_issues = repo.get_issues(state="open")
            for issue in open_issues:
                issue.edit(state="closed")
            return "All open issues closed successfully."
        except Exception as e:
            return str(e)



import logging
from typing import List, Optional
from .stateful_tool_entity import *
from ...agents.software_engineer.software_engineer import SoftwareEngineerAgent
    
class SWEToolEntity(StatefulToolEntity):
    def __init__(self):
        super().__init__("swe_tool.yaml")
        self.tmp_dict = {}
        self.task = SoftwareEngineerAgent()
        self.previous_action = []

    def _call(self, **kwargs):
        if "goto" not in kwargs:
            if self.current_stage.name == self.config.start_stage:
                return {
                    "type": "success",
                    "content": {"message": "swe tool is started"},
                }
            else:
                return {
                    "type": "error",
                    "content": {"message": "please provide `goto` parameter"},
                }

        request_next_stage = kwargs["goto"]
        if request_next_stage not in self.config.all_stages:
            return {
                "type": "error",
                "content": {"message": f"stage {request_next_stage} not found"},
            }
        self.current_stage = self.config.all_stages[request_next_stage]
        

        match self.current_stage.name:
            case "stage_1":
                return self._stage1(kwargs["repo_url"])
            case "stage_2":
                return self._stage2(kwargs["feature"])
            case "stage_3":
                return self._stage3(kwargs["focus_files_name_list"])
            case "stage_4":
                return self._stage4(kwargs["action"],kwargs["plan_idx"],kwargs["focus_file_name"],kwargs["description"])
            case "stage_5":
                return self._stage5()
            case self.config.finish_stage:
                return self._finish()
            case _:
                return {
                    "type": "error",
                    "content": {"message": f"stage {self.current_stage.name} not found"},
                }

    def _stage1(self, repo_url:str):
        self.repo_url = repo_url
        #本地处理repo
        if repo_url.startswith("http"):
            self.task.set_repo_url(repo_url)
        else:
            self.task.set_local_repo(repo_url)
        return {"type": "success", "content": {"message": "stage1 done,get repo_url"}}

    def _stage2(self, feature: str):
        self.feature = feature
        #添加功能说明
        self.task.set_feature_description(feature)
        return {"type": "success", "content": {"message": "stage2 done,get feature"}}

    def _stage3(self,focus_files_name_list: List[str]):
        #锁定文件
        self.task.set_focus_files()
        #选择修改文件
        for focus_files_name in focus_files_name_list:
            self.task.add_focus_file(focus_files_name)
        #plan
        self.task.design_plan()
        plans = self.task.get_plan()
        return {"type": "success", "content": {"result": "stage3 done","plans":plans}}

    def _stage4(self,action:int,plan_idx:int,focus_file_name:str,description:str):
        ADD_PLAN = 0
        REMOVE_PLAN = 1
        MODIFY_PLAN = 2
        
        plans = self.task.get_plan()
        if action == ADD_PLAN:
            plans.append(Plan(action="add", file_path=focus_file_name, description=description))
        elif action == REMOVE_PLAN:
            print("the deleting plan is: ")
            print(plans[plan_idx])
            # plans.pop(plan_idx)
            del plans[plan_idx]
        elif action == MODIFY_PLAN:
            plans[plan_idx].file_path = focus_file_name
            plans[plan_idx].description = description
            plans[plan_idx].action = 'modify'
        self.task.set_plans(plans)
        return {"type": "success", "content": {"result": "stage4 done","plans":plans}}

    def _stage5(self):
        actions = []
        for action in self.task.implement():
            actions.append(action.content)
            self.previous_action.append(action)
        return {"type": "success", "content": {"result": "stage5 done","actions":self.previous_action}}

    #差最后一个步骤，apply还是不apply还是revise
    def _stage6(self,action:int,action_idx:int,revise_comments:str):
        APPLY = 0
        NOT_APPLY = 1
        REVISE = 2
        if action == APPLY:
            file_actions = self.task.get_file_actions()
            self.task.apply_one_file_action(action_idx)
            finish_file_actions = self.task.get_finish_file_actions()
            finish_file_actions.append(file_actions[action_idx])
            self.task.set_finish_file_actions(finish_file_actions)
            del file_actions[action_idx]
            self.task.set_file_actions(file_actions)
        elif action == NOT_APPLY:
            file_actions = self.task.get_file_actions()
            finish_file_actions = self.task.get_finish_file_actions()
            finish_file_actions.append(file_actions[action_idx])
            selftask.set_finish_file_actions(finish_file_actions)
            del file_actions[action_idx]
            selftask.set_file_actions(file_actions)
        elif action == REVISE:
            new_action = task.agent._revise_code(f"{task.previous_action[action_idx]}", revise_comments, action_idx)
            self.previous_action.append(new_action)
            return {"type": "success", "content": {"result": "stage6 done","actions":self.previous_action}}
        return {"type": "success", "content": {"result": "stage6 done"}}
    
    def _finish(self):
        return {"type": "success", "content": {"message": "stateful tool is finished"}}


from typing import Callable
from abc import ABC, abstractmethod
from pydantic import BaseModel, Field


class NodeInput(BaseModel):
    func_name: str
    func_input: BaseModel


class NodeConfig(BaseModel):
    name: str = Field(description="Node 名称")
    description: str = Field(default="", description="Node 描述")
    functions: dict[str, str] = Field(default={}, description="Node 所有功能描述")


class BaseNode(ABC):
    config: NodeConfig
    func_mapping: dict[str, Callable]

    def __init__(self):
        # initialize func_mapping
        self.func_mapping = {}
        avail_funcs = [
            func_name for func_name in dir(self) if not func_name.startswith("_")
        ]
        for func_name in self.config.functions.keys():
            if func_name not in avail_funcs:
                raise Exception(
                    f"Node {self.config.name} does not contain {func_name} method."
                )
            else:
                self.func_mapping[func_name] = getattr(self, func_name)

    def run(self, input: NodeInput):
        if input.func_name not in self.func_mapping.keys():
            raise Exception(
                f"Node {self.config.name} does not contain {input.func_name} method."
            )
        else:
            return self.func_mapping[input.func_name](input.func_input)



import uuid
from collections import deque
from typing import List, Optional
from pydantic import BaseModel, Field

from .assistant import Assistants
from ..nodes.openai.openai import OpenAINode,AsyncOpenAINode
from ..nodes.openai.openai_model import *
from .tools.tools import Tools, Tool
from .config import *
import time
import yaml
import os
import re
import logging
import json


from .prompt.few_shot_cot_tools_choose_prompt import *
from .prompt.parameters_generate_prompt import *
from .prompt.response_generate_prompt import *


def extract_bracket_content(s: str) -> list:
    content = re.findall(r"\[(.*?)\]", s)
    content = [c.replace("'", "") for c in content]
    content = filter(lambda x: x != "", content)
    ret = []
    for item in content:
        if "," in item:
            ret.extend(item.split(","))
        else:
            ret.append(item)
    return ret


class AsyncThreads:
    current_tool: Tool
    chat_node: OpenAINode  # Threads 全局的 OpenAI node，仅用于 chat 交互以及对 tool 执行结果的分析（选择 tool 以及生成参数不使用该 node）

    def __init__(self, config: ThreadsConfig,yaml_file_path:str,threads_yaml_path:Optional[str] = None):
        self.yaml_file_path = yaml_file_path
        self._config = config
        self.current_tool = None
        YamlPathConfig.threads_yaml_path = threads_yaml_path if threads_yaml_path else "threads.yaml"

    @property
    def config(self):
        return self._config

    def set_threads_yaml_path(yaml_path:str):
        # 检查 yaml_path 是否为绝对路径
        if not os.path.isabs(yaml_path):
            # 获取调用此方法的栈帧
            stack = inspect.stack()
            caller_frame = stack[1]
            # 获取调用者的文件路径
            caller_path = caller_frame.filename
            # 获取调用者的目录路径
            caller_dir = os.path.dirname(caller_path)
            # 构建 yaml 文件的绝对路径
            full_yaml_path = os.path.join(caller_dir, yaml_path)
        else:
            full_yaml_path = yaml_path
        # 获取 yaml 文件所在的目录
        yaml_dir = os.path.dirname(full_yaml_path)
        # 如果目录不存在，则创建它
        os.makedirs(yaml_dir, exist_ok=True)
        # 设置 yaml_path
        YamlPathConfig.threads_yaml_path = full_yaml_path

    def save_to_yaml(self):
        # 构建 threads.yaml 文件的绝对路径
        threads_yaml_path = YamlPathConfig.threads_yaml_path
        # 检查文件是否存在，如果不存在，则创建一个空的yaml文件
        if not os.path.exists(threads_yaml_path):
            with open(threads_yaml_path, 'w') as file:
                file.write('')  # 创建一个空文件
        # 使用绝对路径打开 threads.yaml 文件
        with open(threads_yaml_path, "r") as file:
            data = yaml.safe_load(file) or []
        # 查找具有相同 id 的 assistant
        for i, d in enumerate(data):
            if d["id"] == self.config.id:
                # 如果找到了，就更新它
                data[i] = self.config.to_dict()
                break
        else:
            # 如果没有找到，就添加新的 assistant 到列表中
            data.append(self.config.to_dict())
        # 写回 YAML 文件
        with open(threads_yaml_path, "w") as file:
            yaml.dump(data, file)

    @staticmethod
    def create(yaml_file_path:str) -> "Threads":
        # 创建 ThreadsConfig 对象
        config = ThreadsConfig(
            id=str(uuid.uuid4()),
            object="thread",
            created_at=int(time.time()),
            message_history=[],
            metadata={},
        )
       
        # 创建 Threads 对象
        threads = Threads(config,yaml_file_path,YamlPathConfig.threads_yaml_path)

        # 保存到 YAML 文件
        threads.save_to_yaml()

        return threads

    async def run(self, assistant_id: str, input_text: str, **kwargs):
        
        # 使用 from_id 方法获取助手
        assistant = Assistants.from_id(assistant_id)
        tools_list = assistant.get_tools_type_list()
        # 初始化 Tools 对象
        tools = Tools(self.yaml_file_path)
        # 获取 tools 的 summary
        tools_summary = tools.get_tools_list_summary(tools_list)
        # 如果第一次执行或当前的 tool 已执行完毕
        if self.current_tool is None or self.current_tool.has_done():
            # 使用 LLM 选择 tools
            chosen_tools =await self._choose_tools(tools_summary, input_text)
            # TODO: 支持多个 tool 执行
            if len(chosen_tools) == 0:
                logging.warn("No tool is recommended.")

                self.current_tool = None
                # 不使用 Tool, 直接 chat
                res_message = self._chat(input_text, assistant)
            else:
                tool_name = chosen_tools[0]

                # 获取对应的 tool 对象
                self.current_tool = tools.get_tool(tool_name)

        # 判断当前 tool 的执行是否需要 llm 生成参数
        if self.current_tool is not None and self.current_tool.need_llm_generate_parameters():
            # 使用 LLM 生成参数
            parameters = self._generate_parameters(self.current_tool, input_text)
        else:
            parameters = kwargs
            parameters['input_text'] = input_text

        # 执行 tool
        if self.current_tool is not None:
            res_message = self.current_tool.call(**parameters)

        # 根据执行结果，交给 LLM 进行包装
        if self.current_tool is not None and self.current_tool.need_llm_generate_response():
            # 使用 LLM 生成 response
            res_message = self._generate_response(
                self.current_tool, input_text, parameters, res_message, assistant
            )

        #有问题 assistant_message_str = res_message if isinstance(res_message, str) else json.dumps(res_message)
        assistant_message_str = str(res_message)

        self._config.message_history.append(
            [
                MessageRecord(role="user", content=input_text),
                MessageRecord(role="assistant", content=assistant_message_str),
            ]
        )

        return res_message

    async def _chat(
        self, prompt: str, assistant: Assistants, system_message: Optional[str] = None
    ) -> str:
        # 创建一个 OpenAINode 对象
        response_node = AsyncOpenAINode()

        # 使用 assistant 的 description 和 instructions
        description = assistant.description
        instructions = assistant.instructions
        system_prompt = f"""You're an assistant. That's your description.\n{description}\nPlease follow these instructions:\n{instructions}\n """
        response_node.add_system_message(system_prompt)
        if system_message:
            response_node.add_system_message(system_message)

        # 如果 self._config.message_history 里有数据
        if self._config.message_history:
            for record in self._config.message_history:
                user_message = record[0]
                assistant_message = record[1]
                response_node.add_content(user_message.content)
                response_node.add_role(user_message.role)
                response_node.add_content(assistant_message.content)
                response_node.add_role(assistant_message.role)

        message_config = Message(role="user", content=prompt)

        # 创建一个 ChatInput 对象
        chat_config = ChatWithMessageInput(
            message=message_config,
            model="gpt-4-1106-preview",
            append_history=False,
            use_streaming=False,
        )

        response = await response_node.chat_with_message(chat_config)
        response = response.message.content
        return response

    async def _choose_tools(self, tools_summary: dict, input_text: str,instruct:bool = False) -> list[str]:
        
         # 创建一个 OpenAINode 对象
        tools_node = AsyncOpenAINode()
        if instruct:
            tools_choose_prompt = TOOLS_CHOOSE_PROMPT + TOOLS_CHOOSE_EXAMPLE_PROMPT + TOOLS_CHOOSE_HINT +f"""\nInput:\ntools_summary: {tools_summary}\ninput_text: {input_text}\nDispose:"""     

            # 创建一个 ChatInput 对象
            chat_config = OldCompleteInput(
                model="gpt-3.5-turbo-instruct",
                prompt = tools_choose_prompt,
                use_streaming=False
            )

            response = await tools_node.use_old_openai_with_prompt(chat_config)
            response = response.text

        else:
            tools_node.add_system_message(
                TOOLS_CHOOSE_PROMPT + TOOLS_CHOOSE_EXAMPLE_PROMPT + TOOLS_CHOOSE_HINT
            )

            tools_choose_prompt = f"""
    Input:
    tools_summary: {tools_summary}
    input_text: {input_text}
    Dispose:
    """

            message_config = Message(role="user", content=tools_choose_prompt)

            # 创建一个 ChatInput 对象
            chat_config = ChatWithMessageInput(
                message=message_config,
                model="gpt-4-1106-preview",
                append_history=False,
                use_streaming=False,
            )
            # 使用 chat_with_prompt_template 方法进行聊天
            response =await tools_node.chat_with_message(chat_config)
            response = response.message.content  # 现在可以安全地访问 message 属性
            # 使用 chat_with_prompt_template 方法进行聊天
            # response = await tools_node.chat_with_message(chat_config).message.content
        # tools_list = extract_bracket_content(response)
         # 使用正则表达式匹配字典部分
        match = re.search(r'\{.*\}', response, re.DOTALL)
        if match:
            dict_str = match.group()
            # 使用json.loads()函数将字符串转换为字典
            response = json.loads(dict_str)
        else:
            response = json.loads(response)
        tools_list = response['tool']['name']
        return tools_list

    async def _generate_parameters(self, target_tool: Tool, input_text: str,instruct:bool = False) -> dict:
        # 创建一个 OpenAINode 对象
        tools_node = AsyncOpenAINode()
        if instruct:
            parameters_generate_prompt = PARAMETERS_GENERATE_PROMPT + PARAMETERS_GENERATE_EXAMPLE_PROMPT + PARAMETERS_GENERATE_HINT +f"""
    Input:
    tools_name: {target_tool.config.name}
    tools_summary: {target_tool.config.summary}
    input_text: {input_text}
    tool_input_schema: {[parameter.json() for parameter in target_tool.config.parameters]}
    """

            # 创建一个 ChatInput 对象
            chat_config = OldCompleteInput(
                model="gpt-3.5-turbo-instruct",
                prompt = parameters_generate_prompt,
                use_streaming=False
            )
            
            max_attempts = 5
            attempts = 0

            while attempts < max_attempts:
                try:
                    response =await tools_node.use_old_openai_with_prompt(chat_config)
                    response = response.text
                    parameters = json.loads(response)
                    break
                except json.JSONDecodeError:
                    attempts+=1
                    continue
        else:
            tools_node.add_system_message(
                PARAMETERS_GENERATE_PROMPT
                + PARAMETERS_GENERATE_EXAMPLE_PROMPT
                + PARAMETERS_GENERATE_HINT
            )

            parameters_generate_prompt = f"""
    Input:
    tools_name: {target_tool.config.name}
    tools_summary: {target_tool.config.summary}
    input_text: {input_text}
    tool_input_schema: {[parameter.json() for parameter in target_tool.config.parameters]}
    """

            message_config = Message(role="user", content=parameters_generate_prompt)

            # 创建一个 ChatInput 对象
            chat_config = ChatWithMessageInput(
                message=message_config,
                model="gpt-4-1106-preview",
                append_history=False,
                use_streaming=False,
            )

            # 使用 chat_with_prompt_template 方法进行聊天
            max_attempts = 5
            attempts = 0

            while attempts < max_attempts:
                try:
                    response =await tools_node.chat_with_message(chat_config)
                    response = response.message.content
                    parameters = json.loads(response)
                    break
                except json.JSONDecodeError:
                    attempts += 1
                    continue
        return parameters

    async def _generate_response(
        self,
        target_tool: Tool,
        input_text: str,
        tool_input: dict[str, any],
        tool_result: dict[str, any],
        assistant: Assistants,
    ) -> str:
        system_message = (
            RESPONSE_GENERATE_PROMPT
            + RESPONSE_GENERATE_EXAMPLE_PROMPT
            + RESPONSE_GENERATE_HINT
        )
        response_generate_prompt = f"""
Input:
input_text: {input_text}
chosen_tool_info: {target_tool.config.json()}
tool_input: {tool_input}
tool_result: {tool_result}
"""
        return await self._chat(response_generate_prompt, assistant, system_message)


from .code_runner.code_runner import *
from .code_runner.code_runner_model import *

from .openai.openai import *
from .openai.openai_model import *

from .document_loader.document_loader import *
from .document_loader.document_model import *

from .data_analysis.data_analysis import *
from .data_analysis.data_analysis_model import *

from .git.file_io import *
from .git.file_io_model import *
from .git.git_base_node import *
from .git.git_branch import *
from .git.git_commit import *
from .git.git_repo import *



from pydantic import BaseModel, Field
from fastapi import UploadFile, File
from typing import List, Any, Optional
from langchain.docstore.document import Document as LangchainDocument
from pathlib import Path
from datetime import datetime
import hashlib

from ....utils.logger import get_logger
from ...common_models import (
    UserProperties,
    DEFAULT_DOCUMENTS_FOLDER,
    DEFAULT_GIT_FOLDER,
    TIME_STRING_FORMAT,
)


logger = get_logger(__name__)


class UrlDocumentInput(BaseModel):
    url: str = Field(description="Url for generating documents")
    type: str = Field(description="Type of loader used for this url")


class Document(BaseModel):
    user_properties: Optional[UserProperties] = Field(
        UserProperties(), description="User properties."
    )
    file_id: Optional[str] = Field(
        None,
        description="The ID of the uploaded file, generated using the user ID, session ID, and file name.",
    )
    file_path: Optional[Path] = Field(
        None, description="The path to the uploaded file."
    )
    file_name: Optional[str] = Field(
        None,
        description="The name of the uploaded file, including the timestamp of the upload.",
    )
    file_extension: Optional[str] = Field(
        "", description="The extension of the uploaded file."
    )
    file_size: Optional[int] = Field(
        None, description="The size of the uploaded file in bytes."
    )
    creation_time: Optional[str] = Field(
        None,
        description="The timestamp of the upload in the format 'YYYY-MM-DD-HH:MM:SS'.",
    )
    documents: Optional[List[LangchainDocument]] = Field(
        [],
        description="A list of LangchainDocument objects representing the uploaded file.",
    )

    def __init__(self, **kwargs):
        """
        Initializes a new Document object.

        Args:
            input (UploadFile): The uploaded file.
        """
        super().__init__(**kwargs)

    @classmethod
    def create_document_from_file(cls, input: UploadFile, properties: UserProperties):
        """
        Creates a new Document object from an UploadFile input.

        Args:
            input (UploadFile): The uploaded file.

        Returns:
            Document: A new Document object.
        """
        current_time = datetime.now().strftime(TIME_STRING_FORMAT)
        file_name = f"{current_time}-{input.filename}"
        file_path = (
            DEFAULT_DOCUMENTS_FOLDER
            / properties.user_id
            / properties.session_id
            / file_name
        )
        file_path.parent.mkdir(parents=True, exist_ok=True)
        with open(file_path, "wb") as f:
            f.write(input.file.read())

        file_id = hashlib.sha256(
            f"{properties.user_id}-{properties.session_id}-{file_name}".encode()
        ).hexdigest()
        file_extension = file_name.split(".")[-1].lower()
        file_size = input.file._file.tell()
        creation_time = current_time

        return cls(
            user_properties=UserProperties(**properties.dict()),
            file_id=file_id,
            file_path=file_path,
            file_name=file_name,
            file_extension=file_extension,
            file_size=file_size,
            creation_time=creation_time,
        )

    @classmethod
    def create_document_from_url(
        cls, input: UrlDocumentInput, properties: UserProperties
    ):
        current_time = datetime.now().strftime(TIME_STRING_FORMAT)
        file_name = input.url
        file_id = hashlib.sha256(
            f"{properties.user_id}-{properties.session_id}-{file_name}".encode()
        ).hexdigest()
        file_path = None
        file_extension = input.type
        creation_time = current_time
        if input.type == "git":
            file_path = (
                DEFAULT_GIT_FOLDER
                / properties.user_id
                / properties.session_id
                / input.url.split("/")[-1]
            )
            file_path.parent.mkdir(parents=True, exist_ok=True)

        return cls(
            user_properties=UserProperties(**properties.dict()),
            file_id=file_id,
            file_name=file_name,
            file_path=file_path,
            file_extension=file_extension,
            creation_time=creation_time,
        )


class SplitDocumentInput(BaseModel):
    user_properties: Optional[UserProperties] = UserProperties()
    file_id: Optional[str] = None
    chunk_size: Optional[int] = 500
    chunk_overlap: Optional[int] = 0
    document: Optional[Document] = None



from ..base_node import BaseNode, NodeConfig
from .git_base_node import GitBaseNode
from .git_model import (
    GitRepositoryInput,
    GitRemoteRepositoryInput,
    GitConfigInput,
    GitConfigReaderInput,
)

import os
from dotenv import load_dotenv

load_dotenv()  # This will load variables from .env into the environment

GITHUB_TOKEN = os.environ.get("GITHUB_TOKEN")

repository_node_config = {
    "name": "git_repo",
    "description": "A node that implements common repo-level git operations.",
    "functions": {
        "git_init": "Initialize a git repository at given dir path.",
        "git_clone": "Create a new file in the repository.",
        "git_fetch": "Delete a file in the repository.",
        "git_get_config": "Get repo config.",
        "git_config": "Set the config for repo.",
        "git_config_global": "Set global git config for repo",
        "is_dirty": "Return true if repo has untracked file or index has changed.",
    },
}


class GitRepoNode(GitBaseNode):
    config: NodeConfig = NodeConfig(**repository_node_config)

    def __init__(self):
        super().__init__()

    # setters, updaters
    def git_init(self, input: GitRepositoryInput):
        try:
            repo_path = input.path
            self.r.init(repo_path)  # raise NoSuchPathError if path doesn't exist
            return {"status": "Repo initialized successfully."}
        except Exception as e:
            return str(e)

    def git_clone(self, input: GitRemoteRepositoryInput):
        try:
            repo_url = input.url
            repo_path = input.path
            self.r.clone_from(repo_url, repo_path)
            return {"status": "Repo cloned successfully."}
        except Exception as e:
            return str(e)

    def git_fetch(self, input: GitRepositoryInput):
        try:
            repo_path = input.path
            repo = self.g.Repo(repo_path)
            for remote in repo.remotes:
                remote.fetch()
        except Exception as e:
            return str(e)

    def git_get_config(self, input: GitConfigReaderInput):
        try:
            repo_path = input.path
            repo = self.g.Repo(repo_path)
            keys = input.keys
            reader = repo.config_reader(config_level="repository")
            res = []
            for k in keys:
                kk = k.split(".")
                if len(kk) < 2:
                    raise ValueError("Git config key should have section.option format")
                section, option = k
                res.append(reader.get_value(section, option))
            return {"status": "Git local config set successfully."}
        except Exception as e:
            return str(e)

    def git_config(self, input: GitConfigInput):
        try:
            repo_path = input.path
            repo = self.g.Repo(repo_path)
            config = input.config
            writer = repo.config_writer(config_level="repository")
            for k, v in config.items():
                kk = k.split(".")
                if len(kk) < 2:
                    raise ValueError("Git config key should have section.option format")
                section, option = kk
                writer.set_value(section, option, v)
            writer.release()
            return {"status": "Git local config set successfully."}
        except Exception as e:
            return str(e)

    def git_config_global(self, input: GitConfigInput):
        try:
            repo_path = "~"
            repo = self.g.Repo(repo_path)
            config = input.config
            w = repo.config_writer(config_level="global")
            for k, v in config.items():
                kk = k.split(".")
                if len(kk) < 2:
                    raise ValueError("Git config key should have section.option format")
                section, option = kk
                w.set_value(section, option, v)
            w.release()
            return {"status": "Git local config set successfully."}
        except Exception as e:
            return str(e)

    def is_dirty(self, input: GitRepositoryInput):
        try:
            repo_path = input.path
            repo = self.g.Repo(repo_path)
            return {"dirty": repo.is_dirty(untracked_files=True)}
        except Exception as e:
            return str(e)



import os
import subprocess
import logging

from pathlib import Path
from itertools import islice
from typing import Optional
from pydantic import BaseModel, Field

from ...nodes import Document


class FileAction(BaseModel):
    action: str = Field(description="Plan action.")
    file_path: str = Field(description="File path for action.")
    content: Optional[str] = Field(description="Content for action.")

    def __str__(self):
        output = f"""
# {self.file_path}

"""

        if self.content is not None:
            if self.content.startswith("```"):
                output += f"""
{self.content}

"""
            else:
                output += f"""
```python
{self.content}
```

"""
        else:
            output += f"""
**Remove this file.**

"""
        return output


class RepoManager:
    """
    This class is responsible for managing the files of the software engineer.
    It will manage all the file from a single git repo.
    """

    root_path: Optional[Path]
    focus_files: set[str]
    file_tree_str: str
    readme_content: str

    def __init__(self, repo_path: Path):
        self.root_path = repo_path
        self.focus_files = set()
        self.file_tree_str = self._generate_file_tree_str()
        self.readme_content = self._load_readme()

    def add_focus_file(self, file_path):
        """
        Add a file to the interest file set.
        """
        full_path = self.root_path / file_path

        if os.path.exists(full_path):
            if file_path in self.focus_files:
                logging.warning(f"Try to add existing file {file_path}.")
            self.focus_files.add(file_path)
        else:
            raise FileNotFoundError(f"File {full_path} does not exist.")

    def remove_focus_file(self, file_path):
        """
        Remove a file from the interest file set.
        """
        if file_path in self.focus_files:
            self.focus_files.remove(file_path)
        else:
            logging.warning(f"Try to remove unexisting file {file_path}.")

    def set_focus_file(self, focus_files: list[str]):
        """
        Set the focus file.
        """
        self.focus_files.clear()
        for file in focus_files:
            self.add_focus_file(file)

    def get_focus_files_content(self, limit_files: int = 5):
        """
        Get the content of the focus files (only output first 5 files).
        """
        return {
            file_path: file_content
            for file_path, file_content in islice(
                zip(self.focus_files, map(self.get_file_content, self.focus_files)),
                limit_files,
            )
        }

    def get_file_content(self, file_path: str, limit_lines: int = 200):
        """
        Get the content of a file.
        """
        full_path = self.root_path / file_path

        if os.path.exists(full_path):
            with open(full_path, "r") as f:
                lines = []
                line = f.readline()
                while line:
                    lines.append(line)
                    line = f.readline()
                    if len(lines) > limit_lines:
                        lines.append("... (too many lines)")
                        break

                return "\n".join(lines)
        else:
            raise FileNotFoundError(f"File {full_path} does not exist.")

    def apply_file_actions(self, file_actions: list[FileAction]):
        """
        Apply the file actions to the files.
        """
        for file_action in file_actions:
            match file_action.action:
                case "add":
                    self._add_file(file_action.file_path, file_action.content)
                case "remove":
                    self._remove_file(file_action.file_path)
                case "modify":
                    self._modify_file(file_action.file_path, file_action.content)
                case _:
                    raise ValueError(f"Unknown action {file_action.action}.")

        # regenerate file tree
        self.file_tree_str = self._generate_file_tree_str()

    def _add_file(self, file_path: str, content: str):
        """
        Add a file to the repo.
        """
        full_path = self.root_path / file_path

        if os.path.exists(full_path):
            raise FileExistsError(f"File {full_path} already exists.")
        else:
            with open(full_path, "w") as f:
                f.write(content)

    def _remove_file(self, file_path: str):
        """
        Remove a file from the repo.
        """
        full_path = self.root_path / file_path

        if os.path.exists(full_path):
            os.remove(full_path)
        else:
            raise FileNotFoundError(f"File {full_path} does not exist.")

    def _modify_file(self, file_path: str, content: str):
        """
        Modify a file in the repo.
        """
        full_path = self.root_path / file_path

        if os.path.exists(full_path):
            with open(full_path, "w") as f:
                f.write(content)
        else:
            raise FileNotFoundError(f"File {full_path} does not exist.")

    def _generate_file_tree_str(self):
        result = subprocess.run(
            ["tree", "."], stdout=subprocess.PIPE, cwd=self.root_path
        )
        result.check_returncode()
        return result.stdout.decode("utf-8")

    def _load_readme(self):
        readme_file_names = ["README.md", "README", "README.txt"]
        for readme_file_name in readme_file_names:
            readme_path = self.root_path / readme_file_name
            if os.path.exists(readme_path):
                with open(readme_path, "r") as f:
                    return f.read()

        return ""



from pathlib import Path

import pandas as pd
import logging

from .data_scientist_prompt import *
from ..base_agent import BaseAgent, AgentConfig

from ...nodes import (
    OpenAINode,
    ChatInput,
    DataAnalysisNode,
    LoadDataInput,
)

data_scientist_config = {
    "name": "data_scientist",
    "description": "A agent for data scientist.",
}


class DataScientistAgent(BaseAgent):
    config: AgentConfig = AgentConfig(**data_scientist_config)

    def __init__(self):
        self.openai_node = OpenAINode()
        self.data_analysis_node = DataAnalysisNode()

        self._init_openai_node()

    def add_data_file(self, file_path: Path):
        file_name = file_path.name
        file_type = file_path.suffix[1:]

        self.data_analysis_node.load_data(
            input=LoadDataInput(
                name=file_name,
                source_type=file_type,
                source_path=str(file_path),
            )
        )

        logging.warning(f"{file_name} loaded as {file_type}.")
        logging.warning(file_type == "csv" or file_type == "json")
        logging.warning(
            DATA_FILE_PROMPT.format(
                file_name=file_name,
                file_path=file_path,
                n=5,
                content=self.get_data_file_summary(file_name, 5),
            )
        )
        if file_type == "csv" or file_type == "json":
            # set system message about the data file
            self.openai_node.add_system_message(
                DATA_FILE_PROMPT.format(
                    file_name=file_name,
                    file_path=file_path,
                    n=5,
                    content=self.get_data_file_summary(file_name, 5, "markdown"),
                )
            )
            logging.warning(f"openai_history")
            logging.warning(self.openai_node.history)
        else:
            self.openai_node.add_system_message(
                COMMON_FILE_PROMPT.format(
                    file_name=file_name,
                    file_path=file_path,
                )
            )

    def get_data_file_summary(self, file_name: str, head_n: int = 5, output_type: str = "html"):
        if self.data_analysis_node.data.get(file_name) is None:
            return "Data file not found."

        if isinstance(self.data_analysis_node.data[file_name], pd.DataFrame):
            if output_type == "html":
                return self.data_analysis_node.data[file_name].head(head_n).to_html()
            elif output_type == "markdown":
                return self.data_analysis_node.data[file_name].head(head_n).to_markdown()
            else:
                return self.data_analysis_node.data[file_name].head(head_n).to_string()

        with open(self.data_analysis_node.data[file_name], "r") as f:
            content = f.read(300)
            if f.read(1) != "":
                content += "..."

        return content

    def query(self, query: str) -> str:
        resp = self.openai_node.chat(
            input=ChatInput(
                model="gpt-4",
                message_text=QUERY_PROMPT.format(query=query),
                append_history=True,
            )
        )

        code = resp.message.content

        if code.startswith("```python"):
            code = code[10:]

        if code.endswith("```"):
            code = code[:-3]

        return code

    def _init_openai_node(self):
        self.openai_node.add_system_message(
            DS_PROMPT,
        )



import re
import ast
import typeguard

from abc import ABC, abstractmethod, abstractclassmethod
from subprocess import CompletedProcess


class Block(ABC):
    def __init__(self, text: str, title: str):
        self.raw_text = text
        self.title = title

    @abstractmethod
    def parse(self):
        pass

    @abstractmethod
    def content(self):
        pass


class CodeBlock(Block):
    CODE_BLOCK_PATTERN = r"```{lang}.*?\s+(.*?)\s+```"

    def __init__(self, text: str, title: str, lang: str = ""):
        super().__init__(text, title)
        self.lang = lang
        self.code = ""

    def parse(self):
        pattern = self.CODE_BLOCK_PATTERN.format(lang=self.lang)
        match = re.search(pattern, self.raw_text, re.DOTALL)
        if match:
            self.code = match.group(1)
        else:
            raise Exception(
                f"Cannot parse {self.lang} code block in text: \n{self.raw_text}"
            )

    def content(self):
        return self.code


class PythonCodeBlock(CodeBlock):
    def __init__(self, text: str, title: str):
        super().__init__(text, title, lang="python")
        self.pyobj = None

    def parse(self):
        super().parse()
        if not self._is_valid_python(self.code):
            raise Exception(f"Code block is not valid python: {self.code}")
        else:
            try:
                self.pyobj = ast.literal_eval(self.code)
            except Exception as e:
                raise Exception(
                    f"Cannot parse python code block: {self.code}, error: {e}"
                )

    def content(self):
        return self.pyobj

    def _is_valid_python(self, code: str) -> bool:
        try:
            ast.parse(code)
            return True
        except SyntaxError:
            return False


class StringBlock(Block):
    def __init__(self, text: str, title: str):
        super().__init__(text, title)
        self.text = ""

    def parse(self):
        self.text = self.raw_text

    def content(self):
        return self.text


# Output Converter Class:
#


class OuptutParser(ABC):
    @classmethod
    @abstractclassmethod
    def parse_output(cls, input):
        pass

    @classmethod
    @abstractclassmethod
    def parse_output_with_schema(cls, input, schema):
        pass


class LLMOutputParser(OuptutParser):
    """
    input: str -> output: dict[str, any]
    """

    @classmethod
    def parse_output(cls, input: str) -> dict[str, any]:
        block_dict = {}

        block_texts = cls._split_block_text(input)
        for block_text in block_texts:
            block_title, block_content = block_text
            block = cls._parse_block(block_content, block_title)
            block.parse()
            content = block.content()
            block_dict[block_title] = content
        return block_dict

    @classmethod
    def parse_output_with_schema(
        cls, input: str, schema: dict[str, type]
    ) -> dict[str, any]:
        block_dict = {}

        block_texts = cls._split_block_text(input)
        for block_text in block_texts:
            block_title, block_content = block_text

            if block_title in schema.keys():
                block = cls._parse_block(block_content, block_title)
                block.parse()
                content = block.content()
                try:
                    content = typeguard.check_type(content, schema[block_title][0])
                    block_dict[block_title] = content
                except Exception as e:
                    raise Exception(f"Parse {block_title} error: {e}")

        return block_dict

    @classmethod
    def _split_block_text(cls, text: str) -> list[tuple[str, str]]:
        block_texts = text.split("##")
        block_list = []
        for block_text in block_texts:
            block_text = block_text.strip()
            if block_text != "":
                try:
                    block_title, block_content = block_text.split("\n", 1)
                except Exception as e:
                    raise Exception(f"Cannot split block text: {block_text}, error: {e}")
                block_title.strip()
                block_content.strip()
                block_list.append((block_title, block_content))
        return block_list

    @classmethod
    def _parse_block(cls, block_content: str, block_title: str) -> Block:
        if block_content.startswith("```python"):
            return PythonCodeBlock(block_content, block_title)
        else:
            return Block(block_content, block_title)


class RawOutputParser(OuptutParser):
    """
    input: any -> output: any
    """

    @classmethod
    def parse_output(cls, input: any) -> any:
        return input

    @classmethod
    def parse_output_with_schema(
        cls, input: any, _: any
    ) -> any:
        return input

class GitLoaderOutputParser(OuptutParser):
    @classmethod
    def parse_output_with_schema(
        cls, text: str, schema: dict[str, type]
    ) -> dict[str, any]:

        code_content = PythonCodeBlock(text, "code").parse()
        return {"Code Content": code_content}



from pathlib import Path

from ..base_node import BaseNode, NodeConfig
from .git_base_node import GitBaseNode
from .file_io_model import FileInput, DirectoryInput, EditFileInput, EditFileLineInput


file_io_node_config = {
    "name": "file_io",
    "description": "A node that implements common file operations.",
    "functions": {
        "create_file": "Create a file.",
        "create_file_with_content": "Create a file with given content.",
        "delete_file": "Delete a file.",
        "create_directory": "Create a directory.",
        "edit_file": "Edit a file.",
        "edit_file_line": "Edit a line of a file.",
    },
}


class FileIONode(GitBaseNode):
    config: NodeConfig = NodeConfig(**file_io_node_config)

    def create_file(self, input: FileInput):
        file_path = Path(input.path)
        try:
            file_path.parents[0].mkdir(parents=True, exist_ok=True)
            Path(file_path).touch()
            return {"status": f"File {str(file_path)} created successfully."}
        except Exception as e:
            return str(e)

    def create_file_with_content(self, input: EditFileInput):
        file_path = Path(input.path)
        try:
            file_path.parents[0].mkdir(parents=True, exist_ok=True)
            Path(file_path).touch()
            with file_path.open("w", encoding="utf-8") as f:
                f.write(input.content)
            return {
                "status": f"File {str(file_path)} created successfully with given content."
            }
        except Exception as e:
            return str(e)

    def delete_file(self, input: FileInput):
        file_path = Path(input.path)
        try:
            file_path.parents[0].mkdir(parents=True, exist_ok=True)
            Path(file_path).unlink()
            return {"status": f"File {str(file_path)} deleted successfully."}
        except Exception as e:
            return str(e)

    def create_directory(self, input: DirectoryInput):
        file_path = Path(input.path)
        try:
            file_path.mkdir(parents=True, exist_ok=True)
            return {"status": f"Directory {str(file_path)} created successfully."}
        except Exception as e:
            return str(e)

    edit_file = create_file_with_content

    def edit_file_line(self, input: EditFileLineInput):
        file_path = Path(input.path)
        lines = []
        changed = []
        unchanged = []
        try:
            with open(file_path, "r") as file:
                lines = file.readlines()
            for k, v in input.change_list.items():
                if 0 <= k < len(lines):
                    lines[k] = v
                    changed.appned(k)
                else:
                    unchanged.append(k)

            with open(file_path, "w") as file:
                file.writelines(lines)

            return {
                "status": f"File {str(file_path)} changed successfully, lines {str(changed)} changed, lines {str(unchanged)} unchanged."
            }
        except Exception as e:
            return str(e)



from ..base_node import BaseNode, NodeConfig
from .github_node import GithubNode
from .github_model import (
    CreateCommitStatusCheckInput,
    GetCommitDateInput,
)

from datetime import datetime

github_commit_node_config = {
    "name": "github_commit",
    "description": "A node for interacting with GitHub commits.",
    "functions": {
        "create_commit_status_check": "Create a status check for a commit.",
        "get_commit_date": "Get the date of a commit.",
    },
}


class GithubCommitNode(GithubNode):
    config: NodeConfig = NodeConfig(**github_commit_node_config)

    def __init__(self):
        super().__init__()

    def create_commit_status_check(self, input: CreateCommitStatusCheckInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            commit = repo.get_commit(sha=input.sha)
            status = commit.create_status(
                state=input.state,
                target_url=input.target_url,
                description=input.description,
                context=input.context,
            )
            return {"status": "Status check created successfully.", "url": status.url}
        except Exception as e:
            return str(e)

    def get_commit_date(self, input: GetCommitDateInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            commit = repo.get_commit(sha=input.sha)
            date = commit.commit.committer.date
            return date.strftime("%Y-%m-%d %H:%M:%S")
        except Exception as e:
            return str(e)



from .model import *
from .tool_entity import *
from ...nodes.openai.openai import OpenAINode
from ...nodes.openai.openai_model import *
from ...assistant.prompt.few_shot_extract_parametes_from_input import *
from ...assistant.prompt.few_shot_response_prompt import *
import inspect
import json
class StatefulToolEntityConfig(BaseModel):
    start_stage: str = Field(description="The start stage of the tool entity.")
    finish_stage: str = Field(description="The finish stage of the tool entity.")
    all_stages: dict[str, Stage] = Field(description="All stages of the tool entity.")


class StatefulToolEntity(BaseToolEntity, ABC):
    config: StatefulToolEntityConfig
    current_stage: Stage

    def __init__(self, config_file_name: str):
         # 获取调用此方法的栈帧
        stack = inspect.stack()
        # 假设调用者是栈的第二个帧（第一个是当前的 __init__ 方法）
        caller_frame = stack[1]
        # 获取调用者的文件路径
        caller_path = caller_frame.filename
        # 获取调用者的目录路径
        caller_dir = os.path.dirname(caller_path)
        # 构建 openai.yaml 文件的绝对路径
        config_path = os.path.join(caller_dir, config_file_name)
        # 使用绝对路径打开 assistants.yaml 文件
        with open(config_path, "r") as file:
            data = yaml.safe_load(file) or []

        self.config = StatefulToolEntityConfig(**data)
        self.current_stage = self.config.all_stages[self.config.start_stage]

    def is_stateful(self) -> bool:
        return True

    def current_state(self):
        if self.current_stage.name == self.config.finish_stage:
            return State.DONE
        elif self.current_stage.name == self.config.start_stage:
            return State.IDLE
        else:
            return State.RUNNING

    def need_llm_generate_parameters(self) -> bool:
        return self.current_stage.need_llm_generate_parameters

    def need_llm_generate_response(self) -> bool:
        return self.current_stage.need_llm_generate_response
    
    def _get_next_stages_info(self) -> dict[str, list[dict]]:
        return self.current_stage.dict()["next_stage_entry"]

    def call(self, **kwargs):
        if "goto" in kwargs:
            cur_stage_entry = self._get_next_stages_info()
            parameter_info = cur_stage_entry[next(iter(cur_stage_entry))]
            input_text = kwargs['input_text']
            if len(parameter_info)>0:
                #通过input_text和parameter_info去判断是否有值，如果没有则return error 让他重新输入
                parametes = self._extract_parametes(input_text,parameter_info)
                if len(parametes)==0:
                    return{
                        "type": "error",
                        "assistant": {"message": "Please enter the corresponding parameter value correctly"},
                    }
                else:
                    kwargs.update(parametes)
    
        res = self._call(**kwargs)
        # 正常的应该是我这里处理还是上层处理assistant对话
        cur_stage_entry = self._get_next_stages_info()
        if cur_stage_entry:
            parameter_info = cur_stage_entry[next(iter(cur_stage_entry))]
            if len(parameter_info)==0:
                response = f"""{self.current_stage.name} is ready,Enter any response to continue"""
            else:
                response = self._response_chat(parameter_info)
        else:
            response = f"""{self.current_stage.name} is finish,Enter any response to continue"""
    
        
        return {
            **res,
            "next_stages_info": self._get_next_stages_info(),
            "assistant": {"message": f"{response}"},
        }

    def _response_chat(self,parameter_info:list):
        # 创建一个 OpenAINode 对象
        response_node = OpenAINode()
        system_prompt = RESPONSE_PROMPT + RESPONSE_EXAMPLE_PROMPT + RESPONSE_PROMPT_HINT
        response_node.add_system_message(system_prompt)
        prompt = f"""Parameter information (parameter_info):{parameter_info}\nplease output a sentence to user.\nResponse sentence:\n"""
        message_config = Message(role="user", content=prompt)
        # 创建一个 ChatInput 对象
        chat_config = ChatWithMessageInput(
            message=message_config,
            model="gpt-4-1106-preview",
            append_history=False,
            use_streaming=False,
        )
        response = response_node.chat_with_message(chat_config).message.content
        return response
    def _extract_parametes(self,input_text:str,parameter_info:list):
        # 创建一个 OpenAINode 对象
        parametes_node = OpenAINode()
        parametes_node.add_system_message(EXTRACT_PARAMETES_PROMPT+EXTRACT_PARAMETES_EXAMPLE_PROMPT+EXTRACT_PARAMETES_HINT)
        prompt = f"""\nTextual content (input_text):{input_text}\nParameter information (parameter_info):{parameter_info}\nGenerated parameter reference dictionary:\n"""
        message_config = Message(role="user", content=prompt)


        # 创建一个 ChatInput 对象
        chat_config = ChatWithMessageInput(
            message=message_config,
            model="gpt-4-1106-preview",
            append_history=False,
            use_streaming=False,
        )
        
        # response = parametes_node.chat_with_message(chat_config).message.content
        # parametes = json.loads(response)
        
        max_attempts = 5
        attempts = 0
        while attempts < max_attempts:
            try:
                response = parametes_node.chat_with_message(chat_config).message.content
                print(f'response:{response}')
                parametes = json.loads(response)
                break
            except json.JSONDecodeError:
                attempts += 1
                continue

        return parametes

        
    @abstractmethod
    def _call(self, **kwargs):
        pass
        


RESPONSE_GENERATE_PROMPT = """
As a response generator, you'll provide users with suggestions on response generation. Depending on the provided user input (input_text), chosen tool infomation (chosen_tool_info), tool input (tool_input),  tool result (tool_result) , you'll need to follow these steps:

1. Read and understand the chosen tool infomation (chosen_tool_info):
    - Understand the functionality, feature of the chosen tool.

2. Analyze user input and tool input and result:
    - Understand the user's needs or problems.
    - Understand the tool input generate by LLM and tool result generate by tool.

3. Output:
    - Rearrange the tool's result, make it more readable and understandable.
    - Generate the response that need to be generated as string format.

Note that the generation of response should be based on the user's needs and refer to the chosen tool infomation provided. Follow the steps above and make sure to provide accurate response in the output.
"""

RESPONSE_GENERATE_EXAMPLE_PROMPT = """
Here is some examples about response generating:

Input:
input_text: "Tell me the 17th Fibonacci number."
chosen_tool_info: \{
    "name": "code interpreter",
    "summary": "Run the code through code_interpreter and return a dictionary including a log of the code run and whether it was successful or not.",
    ...
\}
tool_input: \{
    "code": "def fibonacci(n):\n    a, b = 0, 1\n    for _ in range(n-1):\n        a, b = b, a + b\n    return a\n\nprint(fibonacci(17))"
\}
tool_result: \{
    "result": "987\n",
    "success": true
\}

Output:
"The 17th Fibonacci number is 987."
"""

RESPONSE_GENERATE_HINT = """
You should only output the string of response, such as "The 17th Fibonacci number is 987.". Do not output any other information and do not contain quotation marks, such as `, \", \' and so on.
"""


TOOLS_CHOOSE_PROMPT = """
As a tool selector, you'll provide users with suggestions on tool selection. Depending on the provided tool summary (tools_summary) and user input (input_text), you'll need to follow these steps:

1. Read and understand the tool summary (tools_summary):
   - Understand the features, suitcases, and limitations of each tool.

2. Analyze User Input (input_text):
   - Understand the user's needs or problems.
   - Identify keywords or phrases to determine which tool best suits the user's needs.

3. Decision-making logic:
   - Recommend a tool if the user's needs correspond to the tool's functionality.
   - If the user's needs are not suitable for any tool, or if the information is not sufficient to make a judgment, no tool is recommended.

4. Output:
   - If a tool is recommended, output the tool name (toolname).
   - If no tool is recommended, the output is empty.

Note that recommendations for tool selection should be based on the user's needs and refer to the tool summary provided. Follow the steps above and make sure to provide accurate tool selection suggestions in the output.
"""

TOOLS_CHOOSE_EXAMPLE_PROMPT = """
Here is some examples about tools choosing:

Input:
tools_summary: {
  "ToolA": "For text analysis and data extraction",
  "ToolB": "For image processing and editing",
  ToolC: For audio editing and processing
}
input_text: "I need to analyze a piece of text to extract key information."

Dispose:
- Analyze the input_text and identify the requirements as "text analytics and data extraction".
- Depending on the tools_summary, ToolA matches this requirement.

Output:
[ToolA]

Input:
tools_summary: {
  "ToolA": "For text analysis and data extraction",
  "ToolB": "For image processing and editing",
  "ToolC": "For text editing and processing"
}
input_text: "I need to analyze the video and tell me how long it is."

Dispose:
- Analyze the input_text and identify the requirement as "Analyze the video and obtain the video duration".
- According to tools_summary, there is no tool that matches this need.

Output:
[]
"""

TOOLS_CHOOSE_HINT = """
You should only output the python list of tool name, such as [ToolA, ToolB, ToolC]. Do not output any other information and do not contain quotation marks, such as `, \", \' and so on.
"""




from .stateful_tool_entity import *


class ExampleStatefulToolEntity(StatefulToolEntity):
    """
    This example tool entity is stateful, and it has 3 inner stages.

    stage1: take integer x as input
    stage2: take integer y as input
    stage3: no input, return x + y
    """

    def __init__(self):
        super().__init__("example_stateful_tool.yaml")

    def _call(self, **kwargs):
        if "goto" not in kwargs:
            if self.current_stage.name == self.config.start_stage:
                return {
                    "type": "success",
                    "content": {"message": "stateful tool is started"},
                }
            else:
                return {
                    "type": "error",
                    "content": {"message": "please provide `goto` parameter"},
                }

        request_next_stage = kwargs["goto"]
        if request_next_stage not in self.config.all_stages:
            return {
                "type": "error",
                "content": {"message": f"stage {request_next_stage} not found"},
            }
        self.current_stage = self.config.all_stages[request_next_stage]

        match self.current_stage.name:
            case "stage_1":
                return self._stage1(kwargs["x"])
            case "stage_2":
                return self._stage2(kwargs["y"])
            case "stage_3":
                return self._stage3()
            case self.config.finish_stage:
                return self._finish()
            case _:
                return {
                    "type": "error",
                    "content": {
                        "message": f"stage {self.current_stage.name} not found"
                    },
                }

    def _stage1(self, x: int):
        self.x = x
        return {"type": "success", "content": {"message": "stage1 done"}}

    def _stage2(self, y: int):
        self.y = y
        return {"type": "success", "content": {"message": "stage2 done"}}

    def _stage3(self):
        return {"type": "success", "content": {"result": self.x + self.y}}

    def _finish(self):
        return {"type": "success", "content": {"message": "stateful tool is finished"}}



PARAMETERS_GENERATE_PROMPT = """
As a tool's parameters generator, you'll generate tool parameters. Depending on the provided tool name (tool_name), tool summary (tool_summary), user input (input_text) and tool input schema (tool_input_schema), you'll need to follow these steps:

1. Read and understand the tool summary (tools_summary):
    - Understand the functionality, features, and limitations of the tool.

2. Analyze User Input (input_text):
    - Understand the user's needs or problems.
    - Identify keywords or phrases to determine which parameters need to be generated.

3. Analyze tool input schema:
    - Understand the schema of the tool input.
    - Identify the parameters that need to be generated.

4. Output:
    - Generate the parameters that need to be generated as json format.
    - If no parameters need to be generated, the output is empty string.

Note that the generation of tool parameters should be based on the user's needs and refer to the tool summary and tool input schema provided. Follow the steps above and make sure to provide accurate tool parameters in the output.
"""

PARAMETERS_GENERATE_EXAMPLE_PROMPT = """
Here is some examples about parameters generating:

Input:
tool_name: "code interpreter"
tool_summary: "Run the code through code_interpreter and return a dictionary including a log of the code run and whether it was successful or not."
input_text: "Tell me the 17th Fibonacci number."
tool_input_schema: [
    \{
    "name": "code",
    "description": "code text that requires code_interpreter to run",
    "required": true,
    "schema": {
        "type": "string"
    }
\}
]

Output:
\{
"code": "def fibonacci(n):\n    a, b = 0, 1\n    for _ in range(n-1):\n        a, b = b, a + b\n    return a\n\nprint(fibonacci(17))",
\}
"""

PARAMETERS_GENERATE_HINT = """
You should only output the json string of parameters, such as {"code": "print('Hello World')"}. Do not output any other information and do not contain quotation marks, such as `, \", \' and so on.
"""


import openai
from openai import OpenAI, AsyncOpenAI
import os
from typing import Optional

import tokentrim as tt

from ..base_node import BaseNode, NodeConfig
from .openai_model import *

import logging

openai_node_config = {
    "name": "openai",
    "description": "A node that interacts with OpenAI's GPT-3 model.",
    "functions": {
        "chat": "Chat with OpenAI's LLM model.",
        "chat_with_prompt_template": "Chat with OpenAI's LLM model with a specific prompt template.",
        "chat_with_message": "Chat with OpenAI's LLM model with a specific message dict.",
    },
}


class OpenAINode(BaseNode):
    config: NodeConfig = NodeConfig(**openai_node_config)

    history: list[dict[str, any]]
    functions: list[dict[str, any]]

    cur_role: Optional[str]
    cur_content: Optional[str]

    def __init__(self):
        super().__init__()

        self.history = []
        self.functions = []

        self.cur_role = None
        self.cur_content = None

    def complete(self, input: CompleteInput):
        """
        Complete with only current history. No extra messages.
        """
        return self._make_completion([], input)

    # TODO: generalize these chat functions
    def chat(self, input: ChatInput):
        """
        Chat with OpenAI's model with simple text.
        """
        return self._make_completion(
            [
                Message(
                    role="user",
                    content=input.message_text,
                )
            ],
            input,
        )

    def chat_with_prompt_template(self, input: ChatWithPromptTemplateInput):
        """
        Chat with OpenAI's model with a specific prompt template.
        """
        return self._make_completion(
            [
                Message(
                    role="user",
                    content=input.prompt_template.format(**input.params),
                )
            ],
            input,
        )

    def chat_with_message(self, input: ChatWithMessageInput):
        """
        Chat with OpenAI's model with a specific message dict.
        """
        return self._make_completion([input.message], input)

    def chat_with_messages(self, input: ChatWithMessagesInput):
        """
        Chat with OpenAI's model with a specific message dict.
        """
        return self._make_completion(input.messages, input)

    def use_old_openai_with_prompt(self, input: OldCompleteInput):
        return self._make_old_completion(input.prompt, input)

    def _make_old_completion(
        self, prompt: str, input: OldCompleteConfig
    ) -> OpenAIOldResp:
        """
        Make a completion with the given messages.
        """

        kwargs = {"model": input.model, "max_tokens": 1096}

        kwargs["prompt"] = prompt
        # set streaming if needed
        if input.use_streaming:
            kwargs["stream"] = True

        # TODO: add exception handling
        try:
            client = OpenAI(api_key=os.getenv("OPENAI_CHAT_API_KEY"))
            response = client.completions.create(**kwargs)
        except Exception as e:
            logging.warn(f"openai_node._make_completion: error occurred: {e}")
            return OpenAIOldResp(
                text=f"Error occurred: {e}",
                finish_reason="error",
            )

        if input.use_streaming:
            # TODO 目前不支持流式处理
            resp = OpenAIOldResp(text="", finish_reason="")
            for completion in response:
                resp.text += completion["choices"][0]["text"]
                if choice.finish_reason:
                    resp.finish_reason = completion["choices"][0]["finish_reason"]
                    break
            return resp

        resp = OpenAIOldResp(**response.choices[0].model_dump())
        return resp

    def _make_completion(
        self, messages: list[Message], input: ChatConfig
    ) -> OpenAIResp | OpenAIStreamingResp:
        """
        Make a completion with the given messages.
        """

        kwargs = {
            "model": input.model,
        }

        cur_messages = []

        # if history is empty, add a default system message
        if len(self.history) == 0:
            cur_messages.append(
                Message(
                    role="system",
                    content="You are a helpful AI assistant. You should answer the user's questions and help them with their tasks.",
                ).dict(exclude_none=True)
            )
        else:
            cur_messages += self.history

        # append history if needed
        if input.append_history:
            for message in messages:
                self.add_single_message(message)

        # add all input messages to argument `messages`
        for message in messages:
            cur_messages.append(message.dict(exclude_none=True))

        kwargs["messages"] = tt.trim(cur_messages, input.model, max_tokens=9999)

        # add function definitions if exists
        if len(self.functions) > 0:
            kwargs["functions"] = self.functions
            kwargs["function_call"] = "auto"

        # set streaming if needed
        if input.use_streaming:
            kwargs["stream"] = True

        # TODO: add exception handling
        try:
            client = OpenAI(api_key=os.getenv("OPENAI_CHAT_API_KEY"))
            response = client.chat.completions.create(**kwargs)
        except Exception as e:
            logging.warn(f"openai_node._make_completion: error occurred: {e}")
            return OpenAIResp(
                message=Message(
                    role="system",
                    content=f"Error occurred: {e}",
                ),
                finish_reason="error",
            )

        if input.use_streaming:
            resp = OpenAIStreamingResp(**response.choices[0].dict())
            if input.append_history:
                self.history.append(resp.delta.dict(exclude_none=True))
            return resp

        resp = OpenAIResp(**response.choices[0].dict())
        if input.append_history:
            self.history.append(resp.message.dict(exclude_none=True))
        return resp

    def add_function(self, func_def: FunctionDefinition):
        self.functions.append(
            func_def.dict()
        )  # redefined dict() doesn't have exclude_none arg

    def add_single_message(self, msg: Message):
        if self.cur_role is not None and self.cur_content is not None:
            self.history.append(
                Message(
                    role=self.cur_role,
                    content=self.cur_content,
                ).dict(exclude_none=True)
            )
            self.cur_role = None
            self.cur_content = None

        self.history.append(msg.dict(exclude_none=True))

    def add_system_message(self, content: str):
        self.add_single_message(
            Message(
                role="system",
                content=content,
            )
        )

    def add_role(self, role: str):
        if self.cur_role is not None and self.cur_content is not None:
            self.add_single_message(
                Message(
                    role=self.cur_role,
                    content=self.cur_content,
                )
            )

        self.cur_role = role

    def add_content(self, content: str):
        if self.cur_content is not None:
            self.cur_content += content
        else:
            self.cur_content = content


class AsyncOpenAINode(BaseNode):
    config: NodeConfig = NodeConfig(**openai_node_config)

    history: list[dict[str, any]]
    functions: list[dict[str, any]]

    cur_role: Optional[str]
    cur_content: Optional[str]

    def __init__(self):
        super().__init__()

        self.history = []
        self.functions = []

        self.cur_role = None
        self.cur_content = None

        openai.api_key = os.getenv("OPENAI_CHAT_API_KEY")
        openai.api_base = os.getenv("OPENAI_CHAT_API_BASE")

    async def complete(self, input: CompleteInput):
        """
        Complete with only current history. No extra messages.
        """
        return await self._make_completion([], input)

    # TODO: generalize these chat functions
    async def chat(self, input: ChatInput):
        """
        Chat with OpenAI's model with simple text.
        """
        return await self._make_completion(
            [
                Message(
                    role="user",
                    content=input.message_text,
                )
            ],
            input,
        )

    async def chat_with_prompt_template(self, input: ChatWithPromptTemplateInput):
        """
        Chat with OpenAI's model with a specific prompt template.
        """
        return await self._make_completion(
            [
                Message(
                    role="user",
                    content=input.prompt_template.format(**input.params),
                )
            ],
            input,
        )

    async def chat_with_message(self, input: ChatWithMessageInput):
        """
        Chat with OpenAI's model with a specific message dict.
        """
        return await self._make_completion([input.message], input)

    async def chat_with_messages(self, input: ChatWithMessagesInput):
        """
        Chat with OpenAI's model with a specific message dict.
        """
        return await self._make_completion(input.messages, input)

    async def use_old_openai_with_prompt(self, input: OldCompleteInput):
        return await self._make_old_completion(input.prompt, input)

    async def _make_old_completion(
        self, prompt: str, input: OldCompleteConfig
    ) -> OpenAIOldResp:
        """
        Make a completion with the given messages.
        """

        kwargs = {"model": input.model, "max_tokens": 1096}

        kwargs["prompt"] = prompt
        # set streaming if needed
        if input.use_streaming:
            kwargs["stream"] = True

        # TODO: add exception handling
        try:
            client = OpenAI(api_key=os.getenv("OPENAI_CHAT_API_KEY"))
            response = client.completions.create(**kwargs)
        except Exception as e:
            logging.warn(f"openai_node._make_completion: error occurred: {e}")
            return OpenAIOldResp(
                text=f"Error occurred: {e}",
                finish_reason="error",
            )

        if input.use_streaming:
            # TODO 目前不支持流式处理
            resp = OpenAIOldResp(text="", finish_reason="")
            for completion in response:
                resp.text += completion["choices"][0]["text"]
                if choice.finish_reason:
                    resp.finish_reason = completion["choices"][0]["finish_reason"]
                    break
            return resp

        resp = OpenAIOldResp(**response.choices[0].model_dump())
        return resp

    async def _make_completion(
        self, messages: list[Message], input: ChatConfig
    ) -> OpenAIResp | OpenAIStreamingResp:
        """
        Make a completion with the given messages.
        """

        kwargs = {
            "model": input.model,
        }

        cur_messages = []

        # if history is empty, add a default system message
        if len(self.history) == 0:
            cur_messages.append(
                Message(
                    role="system",
                    content="You are a helpful AI assistant. You should answer the user's questions and help them with their tasks.",
                ).dict(exclude_none=True)
            )
        else:
            cur_messages += self.history

        # append history if needed
        if input.append_history:
            for message in messages:
                self.add_single_message(message)

        # add all input messages to argument `messages`
        for message in messages:
            cur_messages.append(message.dict(exclude_none=True))

        kwargs["messages"] = tt.trim(cur_messages, input.model, max_tokens=9999)

        # add function definitions if exists
        if len(self.functions) > 0:
            kwargs["functions"] = self.functions
            kwargs["function_call"] = "auto"

        # set streaming if needed
        if input.use_streaming:
            kwargs["stream"] = True

        # TODO: add exception handling
        try:
            client = AsyncOpenAI(api_key=os.getenv("OPENAI_CHAT_API_KEY"))
            response = await client.chat.completions.create(**kwargs)
        except Exception as e:
            return OpenAIResp(
                message=Message(
                    role="system",
                    content=f"Error occurred: {e}",
                ),
                finish_reason="error",
            )

        if input.use_streaming:
            resp = OpenAIStreamingResp(**response.choices[0].dict())
            if input.append_history:
                self.history.append(resp.delta.dict(exclude_none=True))
            return resp

        resp = OpenAIResp(**response.choices[0].dict())
        if input.append_history:
            self.history.append(resp.message.dict(exclude_none=True))
        return resp

    def add_function(self, func_def: FunctionDefinition):
        self.functions.append(
            func_def.dict()
        )  # redefined dict() doesn't have exclude_none arg

    def add_single_message(self, msg: Message):
        if self.cur_role is not None and self.cur_content is not None:
            self.history.append(
                Message(
                    role=self.cur_role,
                    content=self.cur_content,
                ).dict(exclude_none=True)
            )
            self.cur_role = None
            self.cur_content = None

        self.history.append(msg.dict(exclude_none=True))

    def add_system_message(self, content: str):
        self.add_single_message(
            Message(
                role="system",
                content=content,
            )
        )

    def add_role(self, role: str):
        if self.cur_role is not None and self.cur_content is not None:
            self.add_single_message(
                Message(
                    role=self.cur_role,
                    content=self.cur_content,
                )
            )

        self.cur_role = role

    def add_content(self, content: str):
        if self.cur_content is not None:
            self.cur_content += content
        else:
            self.cur_content = content



from pydantic import BaseModel
from pathlib import Path

# Base input for most repository-related functions
class GitRepositoryInput(BaseModel):
    path: str


class GitRemoteRepositoryInput(BaseModel):
    url: str
    path: Path


class GitConfigInput(GitRepositoryInput):
    config: dict[str, str]  # key-value pair of config strings


# key shape: section.key
# value shape: value
# use this config as if you're running the command git config section.key "value"


class GitConfigReaderInput(GitRepositoryInput):
    keys: list[str]


class GitFileInput(BaseModel):
    path: str


class GitAddFileInput(GitRepositoryInput):
    paths: list[str]


class GitCommitInput(GitRepositoryInput):
    message: str
    author: str


class GitResetInput(GitRepositoryInput):
    paths: list[str]


class GitBranchInput(GitRepositoryInput):
    branch_name: str


class GitLogInput(GitRepositoryInput):
    format_options: list[str]

class GitPullInput(GitRepositoryInput):
    remote: str

class GitPushInput(GitRepositoryInput):
    remote: str



from pydantic import BaseModel
from pathlib import Path

from typing import Optional
from enum import Enum
import json

DEFAULT_DOCUMENTS_FOLDER = Path("src/data/documents/")
DEFAULT_GIT_FOLDER = Path("src/data/git/")
DEFAULT_DATA_ANALYSIS_FOLDER = Path("src/data/data_analysis/")
DEFAULT_USER_ID = "admin"
DEFAULT_SESSION_ID = "admin_session"
TIME_STRING_FORMAT = "%Y-%m-%d-%H:%M:%S"


class RedisKeyType(Enum):
    """
    An enumeration for Redis key types.
    """

    DOCUMENTS = "documents"
    VECTORSTORE = "vectorstore"


class UserProperties(BaseModel):
    user_id: Optional[str] = DEFAULT_USER_ID
    session_id: Optional[str] = DEFAULT_SESSION_ID

    def generate_redis_key_with_type(self, key_type: RedisKeyType):
        return f"{self.user_id}:{self.session_id}:{key_type.value}"

    @classmethod
    def __get_validators__(cls):
        yield cls.validate_to_json

    @classmethod
    def validate_to_json(cls, value):
        if isinstance(value, str):
            return cls(**json.loads(value))
        return value



from git import Commit

from ..base_node import BaseNode, NodeConfig
from .git_base_node import GitBaseNode
from .git_model import (
    GitRepositoryInput,
    GitBranchInput,
    GitPullInput,
    GitPushInput,
)

repository_node_config = {
    "name": "git_branch_node",
    "description": "A node that implements common git branch operations.",
    "functions": {
        "get_git_branch_list": "Add specific file(s) to git index.",
        "git_log_raw": "Get raw output string of git log.",
        "git_log": "Get list of commit info.",
        "git_log_commit_topk": "Get list of commit info, max traceback topk.",
        "git_switch_branch": "Switch to another branch.",
        "git_pull": "Git pull.",
        "git_push": "Git push.",
        "git_pull_remote": "Git pull with specified remote",
        "git_push_remote": "Git push with specified remote",
    },
}


def formatGitCommit(commit: Commit):
    return {
        "repo": str(commit.repo),
        "hexsha": commit.hexsha,
        "author": commit.author,
        "message": commit.message,
        "summary": commit.summary,  # first line of message
        "datetime": commit.committed_datetime,
    }


class GitBranchNode(GitBaseNode):
    config: NodeConfig = NodeConfig(**repository_node_config)

    def get_git_branch_list(self, input: GitRepositoryInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            repo_heads_names = [h.name for h in repo.heads]
            return {"branch": repo_heads_names}
        except Exception as e:
            return str(e)

    def git_switch_branch(self, input: GitBranchInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            repo.git.checkout(input.branch_name)
            return {"status": f"Successfully checkout git branch {input.branch_name}."}
        except Exception as e:
            return str(e)

    def git_log_raw(self, input: GitRepositoryInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            log_content = repo.git.log()
            return {"log": log_content}
        except Exception as e:
            return str(e)

    def git_log(self, input: GitRepositoryInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            log_content = repo.git.log()
            return {"log": log_content}
        except Exception as e:
            return str(e)

    def git_log_commit_topk(self, input: GitRepositoryInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            log_content = repo.git.log()
            return {"log": log_content}
        except Exception as e:
            return str(e)

    def git_pull(self, input: GitPullInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            result = repo.git.pull()
            return {"status": "Git pull success", "result": result}
        except Exception as e:
            return str(e)

    def git_pull_remote(self, input: GitPullInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            result = repo.git.push(input.remote)
            return {"status": "Git pull success", "result": result}
        except Exception as e:
            return str(e)

    def git_push(self, input: GitPushInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            result = repo.git.push()
            return {"status": "Git pull success", "result": result}
        except Exception as e:
            return str(e)

    def git_push_remote(self, input: GitPushInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            result = repo.git.push(input.remote)
            return {"status": "Git pull success", "result": result}
        except Exception as e:
            return str(e)



from pydantic import BaseModel


class FileInput(BaseModel):
    path: str


class DirectoryInput(BaseModel):
    path: str


# replace the entire file content with the new content
class EditFileInput(FileInput):
    content: str


# change file content at specific line int -> str
class EditFileLineInput(FileInput):
    change_list: list[int, str]



from ..base_node import BaseNode, NodeConfig
from .github_node import GithubNode
from .github_model import (
    SearchCodeInput,
    SearchCommitsInput,
    SearchIssuesAndPRsInput,
    SearchLabelsInput,
    SearchRepositoriesInput,
    SearchTopicsInput,
    SearchUsersInput,
)

github_search_node_config = {
    "name": "github_search",
    "description": "A node for searching various entities on GitHub.",
    "functions": {
        "search_code": "Search code.",
        "search_commits": "Search commits.",
        "search_issues_and_prs": "Search issues and pull requests.",
        "search_labels": "Search labels.",
        "search_repositories": "Search repositories.",
        "search_topics": "Search topics.",
        "search_users": "Search users.",
    },
}


class GithubSearchNode(GithubNode):
    config: NodeConfig = NodeConfig(**github_search_node_config)

    def __init__(self):
        super().__init__()

    def search_code(self, input: SearchCodeInput):
        try:
            result = self.g.search_code(
                query=input.query, sort=input.sort, order=input.order
            )
            return [{"name": item.name, "path": item.path} for item in result]
        except Exception as e:
            return str(e)

    def search_commits(self, input: SearchCommitsInput):
        try:
            result = self.g.search_commits(
                query=input.query, sort=input.sort, order=input.order
            )
            return [
                {"sha": item.sha, "message": item.commit.message} for item in result
            ]
        except Exception as e:
            return str(e)

    def search_issues_and_prs(self, input: SearchIssuesAndPRsInput):
        try:
            result = self.g.search_issues(
                query=input.query, sort=input.sort, order=input.order
            )
            return [{"number": item.number, "title": item.title} for item in result]
        except Exception as e:
            return str(e)

    def search_labels(self, input: SearchLabelsInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            labels = repo.get_labels()
            return [
                {"id": label.id, "name": label.name}
                for label in labels
                if input.query.lower() in label.name.lower()
            ]
        except Exception as e:
            return str(e)

    def search_repositories(self, input: SearchRepositoriesInput):
        try:
            result = self.g.search_repositories(
                query=input.query, sort=input.sort, order=input.order
            )
            return [{"full_name": repo.full_name} for repo in result]
        except Exception as e:
            return str(e)

    def search_topics(self, input: SearchTopicsInput):
        try:
            result = self.g.search_topics(
                query=input.query, sort=input.sort, order=input.order
            )
            return [{"name": topic.name} for topic in result]
        except Exception as e:
            return str(e)

    def search_users(self, input: SearchUsersInput):
        try:
            result = self.g.search_users(
                query=input.query, sort=input.sort, order=input.order
            )
            return [{"login": user.login, "name": user.name} for user in result]
        except Exception as e:
            return str(e)



from src.core.nodes.base_node import BaseNode, NodeConfig
from .github_node import GithubNode
from .github_model import (
    RepositoryInput,
    GetMilestoneInput,
    CreateMilestoneInput,
    CreateMilestoneWithDetailsInput,
)

github_milestone_node_config = {
    "name": "github_milestone",
    "description": "A node for interacting with GitHub milestones.",
    "functions": {
        "get_milestone_list": "Get list of milestones.",
        "get_milestone": "Get a specific milestone.",
        "create_milestone": "Create a new milestone.",
        "create_milestone_with_details": "Create a milestone with state and description.",
    },
}


class GithubMilestoneNode(GithubNode):
    config: NodeConfig = NodeConfig(**github_milestone_node_config)

    def __init__(self):
        super().__init__()

    def get_milestone_list(self, input: RepositoryInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            milestones = repo.get_milestones()
            return [{"number": m.number, "title": m.title} for m in milestones]
        except Exception as e:
            return str(e)

    def get_milestone(self, input: GetMilestoneInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            milestone = repo.get_milestone(input.milestone_number)
            return {
                "number": milestone.number,
                "title": milestone.title,
                "description": milestone.description,
                "state": milestone.state,
            }
        except Exception as e:
            return str(e)

    def create_milestone(self, input: CreateMilestoneInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            milestone = repo.create_milestone(title=input.title)
            return milestone.number
        except Exception as e:
            return str(e)

    def create_milestone_with_details(self, input: CreateMilestoneWithDetailsInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            milestone = repo.create_milestone(
                title=input.title, state=input.state, description=input.description
            )
            return milestone.number
        except Exception as e:
            return str(e)



from ..base_node import BaseNode, NodeConfig
from .github_node import GithubNode
from .github_model import GetUserRepositoriesInput, InviteUserToOrgInput

github_user_node_config = {
    "name": "github_user",
    "description": "A node for interacting with GitHub users.",
    "functions": {
        "get_user_repositories": "Returns the repositories of a user.",
        "invite_user_to_org": "Invites a user to an organization.",
    },
}


class GithubUserNode(GithubNode):
    config: NodeConfig = NodeConfig(**github_user_node_config)

    def __init__(self):
        super().__init__()

    def get_user_repositories(self, input: GetUserRepositoriesInput):
        try:
            user = self.g.get_user(input.username)
            repos = user.get_repos()
            return [{"id": repo.id, "name": repo.name} for repo in repos]
        except Exception as e:
            return str(e)

    def invite_user_to_org(self, input: InviteUserToOrgInput):
        try:
            org = self.g.get_organization(input.org_name)
            org.invite_user(user=input.username)
            return f"User {input.username} invited to organization {input.org_name}."
        except Exception as e:
            return str(e)



from github import Github
from ..base_node import BaseNode
import os

class GithubNode(BaseNode):
    def __init__(self):
        self.token = os.environ.get("GITHUB_TOKEN")  # Retrieving the token from the environment
        if not self.token:
            raise ValueError("GITHUB_TOKEN is not set in the environment.")
        self.g = Github(self.token)  # Initializing the GitHub instance with the token
        super().__init__()



from pathlib import Path
import os
from dotenv import load_dotenv

from .git_base_node import GitBaseNode
from .git_model import (
    GitAddFileInput,
    GitRepositoryInput,
    GitCommitInput,
)
from ..base_node import BaseNode, NodeConfig

load_dotenv()  # This will load variables from .env into the environment

repository_node_config = {
    "name": "git_commit",
    "description": "A node that implements common git operations related to git commit.",
    "functions": {
        "git_add_files": "Add specific file(s) to git index.",
        "git_add_all": "Add all files in directory.",
        "git_commit": "Make a commit with message in the repository.",
        "git_commit_with_author": "Make a commit with message and given author in the repository.",
        "git_add_commit": "Add files and commit (shorthand of git add -A and git commit).",
        "git_reset_all": "Restore the index, undo git add.",
        "get_untracked_files": "Get list of untracked files of current repo",
    },
}


class GitCommitNode(GitBaseNode):
    config: NodeConfig = NodeConfig(**repository_node_config)

    def git_add_files(self, input: GitAddFileInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            for path in input.paths:
                repo.git.add(path)
            return {
                "status": f"Files {input.paths} are added successfully to index of repo {input.path}."
            }
        except Exception as e:
            return str(e)

    def git_add_all(self, input: GitRepositoryInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            repo.git.add(all=True)
            return {
                "status": f"Files {input.paths} are added successfully to index of repo {input.path}."
            }
        except Exception as e:
            return str(e)

    def get_untracked_files(self, input: GitRepositoryInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            return {"untracked": repo.untracked_files}
        except Exception as e:
            return str(e)

    def git_commit(self, input: GitCommitInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            result = repo.git.commit("-m", input.message)
            return {"status": "Git commit success", "message": result.message}
        except Exception as e:
            return str(e)

    def git_commit_with_author(self, input: GitCommitInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            result = repo.git.commit("-m", input.message, author=input.author)
            return {"status": "Git commit success", "message": result.message}
        except Exception as e:
            return str(e)

    def git_add_commit(self, input: GitCommitInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            repo.git.add(all=True)
            result = repo.git.commit("-m", input.message)
            return {"status": "Git commit success", "message": result.message}
        except Exception as e:
            return str(e)

    def git_reset_all(self, input: GitRepositoryInput):
        repo_path = input.path
        try:
            repo = self.g.Repo(repo_path)
            repo.git.reset()
            return {"status": f"Successfully reset git index"}
        except Exception as e:
            return str(e)

    def git_reset_files(self, input: GitRepositoryInput):
        repo_path = input.path
        done_reset = []
        try:
            repo = self.g.Repo(repo_path)
            for path in input.paths:
                repo.git.reset(path)
                done_reset.append(path)
            return {
                "status": f"Successfully reset git index",
                "reset_files": done_reset,
            }
        except Exception as e:
            return str(e)



from enum import Enum
from typing import Optional
from pydantic import BaseModel, Field


class OpenAIModel(str, Enum):
    GPT_4 = "gpt-4"
    GPT_4_0613 = "gpt-4-0613"
    GPT_4_32K = "gpt-4-32k"
    GPT_4_32K_0613 = "gpt-4-32k-0613"
    GPT_4_1106_PREVIEW= "gpt-4-1106-preview"
    GPT_3_5_TURBO = "gpt-3.5-turbo"
    GPT_3_5_TURBO_16K = "gpt-3.5-turbo-16k"
    GPT_3_5_TURBO_INSTRUCT = "gpt-3.5-turbo-instruct"
    GPT_3_5_TURBO_0613 = "gpt-3.5-turbo-0613"
    GPT_3_5_TURBO_16K_0613 = "gpt-3.5-turbo-16k-0613"
    TEXT_MODERATION_LATEST = "text-moderation-latest"
    TEXT_MODERATION_STABLE = "text-moderation-stable"


class Message(BaseModel):
    """
    Message for OpenAI chat.
    """

    class FunctionCall(BaseModel):
        """
        The name and arguments of a function that should be called, as generated by the model.
        """

        name: str = Field(description="Name of the function to call.")
        arguments: str = Field(
            description="Arguments to call the function with, as generated by the model in JSON format."
        )

    role: Optional[str] = Field(description="Role of the message.")
    name: Optional[str] = Field(description="Name of the message.")
    content: Optional[str] = Field(description="Content of the message.")
    function_call: Optional[FunctionCall] = Field(
        description="Function call of the message."
    )


class FunctionDefinition(BaseModel):
    """
    Function definition for OpenAI function call.
    """

    class FunctionParameter(BaseModel):
        """
        Parameter for OpenAI function call.
        """

        name: str = Field(description="Name of the parameter.")
        description: str = Field(description="Description of the parameter.")
        required: bool = Field(
            default=True, description="Whether the parameter is required or not."
        )

    name: str = Field("Name of the function.")
    description: str = Field("Description of the function.")
    parameters: list[FunctionParameter] = Field("Parameters of the function.")

    def dict(self) -> dict[str, any]:
        return {
            "name": self.name,
            "description": self.description,
            "parameters": {
                "type": "object",
                "properties": {
                    param.name: {
                        "type": "string",
                        "description": param.description,
                    }
                    for param in self.parameters
                },
                "required": [param.name for param in self.parameters if param.required],
            },
        }


class ChatConfig(BaseModel):
    model: OpenAIModel = Field(description="Model to use for chat.")
    use_streaming: bool = Field(
        default=False, description="Whether to use streaming output or not."
    )
    append_history: bool = Field(
        default=False, description="Whether to append this message to history."
    )

class OldCompleteConfig(BaseModel):
    model: OpenAIModel = Field(description="Model to use for chat.")
    use_streaming: bool = Field(
        default=False, description="Whether to use streaming output or not."
    )
    

class OldCompleteInput(OldCompleteConfig):
    prompt:str = Field(description="prompt to use for model.")

class CompleteInput(ChatConfig):
    pass


class ChatInput(ChatConfig):
    message_text: str = Field(description="Message to chat.")


class ChatWithPromptTemplateInput(ChatConfig):
    prompt_template: str = Field(description="Prompt template for chat.")
    params: dict[str, str] = Field(description="Parameters for prompt template.")


class ChatWithMessageInput(ChatConfig):
    message: Message = Field(description="Message to chat.")


class ChatWithMessagesInput(ChatConfig):
    messages: list[Message] = Field(description="Messages to chat.")


class OpenAIResp(BaseModel):
    """
    Response from OpenAI with no streaming. It's a chat completion object.
    """

    message: Message = Field(description="Response message")
    finish_reason: str = Field(description="Finish reason.")


# TODO: support streaming response
class OpenAIStreamingResp(BaseModel):
    """
    Response from OpenAI with streaming.
    """

    delta: Message = Field(description="Delta message")
    finish_reason: str = Field(description="Finish reason.")

class OpenAIOldResp(BaseModel):
    """
    Response from OpenAI with no streaming. It's a chat completion object.
    """
    text: str = Field(description="Response message")
    finish_reason: str = Field(description="Finish reason.")



DS_PROMPT = """
You are a data scientist working on a series of data files. 
You will be asked to write code to process the data in the files and answer questions about the data.

You should:
1.  Comprehend the user's requirements carefully & to the letter.
2.  When user ask a question, write a python code to answer the query. Your code should output answer to STDOUT. 
    Your output should only contain the python code content, do not add any other information in your output.
    Do not contain quotation marks (\", \"\"\", \', etc.) at the beginning and the end of the file content.
    Do not contain markdown code block syntax (```, ~~~, etc.) at the beginning and the end of the file content.
"""

DATA_FILE_PROMPT = """
User add a data file {file_name} at {file_path}.
The header {n} rows of the file (dataframe) is as follow:

{content}

Please take care of the data schema and the data type of each column.
"""

COMMON_FILE_PROMPT = """
User add a data file {file_name} at {file_path}.
"""

QUERY_PROMPT = """
User ask a query: {query}.

Please write the code to answer the query.

Please write a python code for me. The output should only contain file content.


[NOTE]
***Your code should output answer to STDOUT. (IMPORTANT: always use python `print` function)***
DO NOT CONTAIN QUOTATION MARKS (\", \"\"\", \', etc.) AT THE BEGINNING AND THE END OF THE FILE CONTENT.
YOUR OUTPUT SHOULD JUST USE `\n` FOR LINE BREAK, DO NOT USE `\\n`.
DO NOT INCLUDE ANY OTHER INFORMATION EXCEPT FOR THE FILE CONTENT IN YOUR OUTPUT.
"""




import os
import shutil
import time
import uuid
from pathlib import Path
from typing import Any, List
from typing import IO

FILE_CONNECTOR_TMP_STORAGE_PATH = os.environ.get(
    "FILE_CONNECTOR_TMP_STORAGE_PATH", "/knowledge_base"
)
_FILE_AGE_CLEANUP_THRESHOLD_HOURS = 24 * 7  # 1 week
_VALID_FILE_EXTENSIONS = [".txt", ".zip", ".pdf"]


def get_file_ext(file_path_or_name: str | Path) -> str:
    _, extension = os.path.splitext(file_path_or_name)
    return extension


def check_file_ext_is_valid(ext: str) -> bool:
    return ext in _VALID_FILE_EXTENSIONS


def write_temp_files(
    user_id: str,
    files: list[tuple[str, IO[Any]]],
    base_path: Path | str = FILE_CONNECTOR_TMP_STORAGE_PATH,
) -> list[str]:
    """Writes temporary files to disk and returns their paths

    NOTE: need to pass in (file_name, File) tuples since FastAPI's `UploadFile` class
    exposed SpooledTemporaryFile does not include a name.
    """
    file_location = Path(base_path) / user_id / str(uuid.uuid4())
    os.makedirs(file_location, exist_ok=True)

    file_paths: list[str] = []
    for file_name, file in files:
        extension = get_file_ext(file_name)
        if not check_file_ext_is_valid(extension):
            raise ValueError(
                f"Invalid file extension for file: '{file_name}'. Must be one of {_VALID_FILE_EXTENSIONS}"
            )

        file_path = file_location / file_name
        with open(file_path, "wb") as buffer:
            # copy file content from uploaded file to the newly created file
            shutil.copyfileobj(file, buffer)

        file_paths.append(str(file_path.absolute()))

    return file_paths


def file_age_in_hours(filepath: str | Path) -> float:
    return (time.time() - os.path.getmtime(filepath)) / (60 * 60)


def clean_old_temp_files(
    files_path: List[str],
) -> None:
    # os.makedirs(Path(base_path) / user_id, exist_ok=True)
    for file in files_path:
        # if file_age_in_hours(Path(base_path) / file) > age_threshold_in_hours:
        #     os.remove(Path(base_path) / user_id / file)
        os.remove(file)



import uuid
from collections import deque
from typing import List, Optional
from pydantic import BaseModel, Field
import time
import yaml
import os
import re
import logging
import json

class MessageRecord(BaseModel):
    role: str = Field(description="角色")
    content: str = Field(description="内容")

class YamlPathConfig:
    threads_yaml_path = 'threads.yaml'
    assistants_yaml_path = 'assistants.yaml'

class AssistantConfig(BaseModel):
    id: str = Field(description="助手 ID")
    object: str = Field(default="assistant", description="对象类型")
    created_at: int = Field(description="创建时间")
    name: str = Field(description="助手名称")
    description: Optional[str] = Field(default=None, description="助手描述")
    model: str = Field(description="模型")
    instructions: str = Field(description="指令")
    tools: list[dict] = Field(description="工具")
    file_ids: list[str] = Field(default=[], description="文件 ID")
    metadata: dict = Field(default={}, description="元数据")

class ThreadsConfig(BaseModel):
    id: str = Field(description="线程 ID")
    object: str = Field(default="thread", description="对象类型")
    created_at: int = Field(description="创建时间")
    assistant_id: Optional[str] = Field(description="助手 ID")
    message_history: deque[List[MessageRecord]] = Field(
        deque(maxlen=10), description="消息"
    )
    metadata: dict = Field(default={}, description="元数据")

    def to_dict(self):
        # Convert the ThreadsConfig object to a dictionary
        data = self.__dict__.copy()
        # Convert the deque to a list
        data["message_history"] = list(data["message_history"])
        return data

    @classmethod
    def from_dict(cls, data):
        # Convert the list back to a deque
        data["message_history"] = deque(data["message_history"], maxlen=10)
        return cls(**data)


import inquirer



def user_input(message: str, prefix=">>> ", arg_type=str):
    questions = inquirer.Text("_arg", message=prefix + message)
    return arg_type(inquirer.prompt([questions])["_arg"])


def user_confirm(message: str, prefix=">>> ", default=False) -> bool:
    questions = inquirer.Confirm("_arg", message=prefix + message, default=default)
    return inquirer.prompt([questions])["_arg"]


def user_checkbox(message: str, choices: list, prefix=">>> ") -> list:
    questions = inquirer.Checkbox("_arg", message=prefix + message, choices=choices)
    return inquirer.prompt([questions])["_arg"]

def user_list(message: str, choices: list, prefix=">>> "):
    questions = inquirer.List("_arg", message=prefix + message, choices=choices)
    return inquirer.prompt([questions])["_arg"]


from typing import Optional
from pydantic import BaseModel, Field


class RunCodeInput(BaseModel):
    code: str = Field(description="Python code to be executed.")


class RunCodeFromFileInput(BaseModel):
    working_dir: str = Field(description="Working directory for the code.")
    file_path: str = Field(description="Entry python file to be executed.")

    kwargs: Optional[dict] = Field(default={}, description="Keyword arguments.")



import uuid
import time
import os
import yaml
from typing import Optional

import inspect
from .config import *
from typing import Dict,Any
class Assistants():
    def __init__(self, config,yaml_path:Optional[str] = None):
        self.config = config
        YamlPathConfig.assistants_yaml_path = yaml_path if yaml_path else 'assistants.yaml'
    
    def set_assistants_yaml_path(yaml_path: str):
        # 检查 yaml_path 是否为绝对路径
        if not os.path.isabs(yaml_path):
            # 获取调用此方法的栈帧
            stack = inspect.stack()
            caller_frame = stack[1]
            # 获取调用者的文件路径
            caller_path = caller_frame.filename
            # 获取调用者的目录路径
            caller_dir = os.path.dirname(caller_path)
            # 构建 yaml 文件的绝对路径
            full_yaml_path = os.path.join(caller_dir, yaml_path)
        else:
            full_yaml_path = yaml_path

        # 获取 yaml 文件所在的目录
        yaml_dir = os.path.dirname(full_yaml_path)
        # 如果目录不存在，则创建它
        os.makedirs(yaml_dir, exist_ok=True)
        # 设置 assistants_yaml_path
        YamlPathConfig.assistants_yaml_path = full_yaml_path

    def save_to_yaml(self):
        # 构建 assistants.yaml 文件的绝对路径
        assistants_yaml_path = YamlPathConfig.assistants_yaml_path
        # 检查文件是否存在，如果不存在，则创建一个空的yaml文件
        if not os.path.exists(assistants_yaml_path):
            with open(assistants_yaml_path, 'w') as file:
                file.write('')  # 创建一个空文件
        # 使用绝对路径打开 assistants.yaml 文件
        with open(assistants_yaml_path, 'r') as file:
            data = yaml.safe_load(file) or []
        # 查找具有相同 id 的 assistant
        for i, d in enumerate(data):
            if d['id'] == self.config.id:
                # 如果找到了，就更新它
                data[i] = self.config.__dict__
                break
        else:
            # 如果没有找到，就添加新的 assistant 到列表中
            data.append(self.config.__dict__)
        # 写回 YAML 文件
        with open(assistants_yaml_path, 'w') as file:
            yaml.dump(data, file)

    @property
    def id(self):
        return self.config.id

    @property
    def name(self):
        return self.config.name

    @name.setter
    def name(self, value):
        self.config.name = value
        self.save_to_yaml()  # 更新 YAML 文件

    @property
    def instructions(self):
        return self.config.instructions

    @instructions.setter
    def instructions(self, value):
        self.config.instructions = value

    @property
    def description(self):
        return self.config.description

    @description.setter
    def description(self, value):
        self.config.description = value

    @property
    def tools(self):
        return self.config.tools

    @tools.setter
    def tools(self, value):
        self.config.tools = value
        self.save_to_yaml()  # 更新 YAML 文件

    @property
    def model(self):
        return self.config.model

    @model.setter
    def model(self, value):
        self.config.model = value
        self.save_to_yaml()  # 更新 YAML 文件

    def get_tools_type_list(self):
        return [tool['type'] for tool in self.config.tools]

    @staticmethod
    def create(name: str = None, instructions: str = None, tools: list[dict] = [{'type':''}], model: str = 'gpt-4', description: str = None, file_ids: list = None) -> 'Assistants':
        # 创建配置和 Assistants 对象
        config = AssistantConfig(
            id=str(uuid.uuid4()),
            created_at=int(time.time()),
            name=name,
            description=description,
            instructions=instructions,
            tools=tools,
            model=model,
            file_ids=file_ids if file_ids is not None else [],
        )
        assistant = Assistants(config,YamlPathConfig.assistants_yaml_path)
        assistant.save_to_yaml()  # 保存到 YAML 文件
        return assistant
    
    @staticmethod
    def get_all_assistants() -> List[Dict[str, Any]]:
        """
        读取 YAML 文件并返回所有 assistants 的信息列表。
        """
        # 确保 YAML 文件路径已经被设置
        if not YamlPathConfig.assistants_yaml_path or not os.path.isfile(YamlPathConfig.assistants_yaml_path):
            raise FileNotFoundError("The assistants YAML file path is not set or the file does not exist.")

        # 读取 YAML 文件
        with open(YamlPathConfig.assistants_yaml_path, 'r') as file:
            assistants_data = yaml.safe_load(file) or []

        # 返回 assistants 数据列表
        return assistants_data
    
    @classmethod
    def from_id(cls, id: str) -> 'Assistants':
        # 使用传入的 yaml_path 参数打开 YAML 文件
        with open(YamlPathConfig.assistants_yaml_path, 'r') as file:
            data = yaml.safe_load(file) or []
        # 查找具有相同 id 的配置
        for d in data:
            if d['id'] == id:
                # 如果找到了，就用这个配置创建一个新的 Assistants 对象
                config = AssistantConfig(**d)
                return cls(config, YamlPathConfig.assistants_yaml_path)  # 使用传入的 yaml_path 创建 Assistants 实例
        # 如果没有找到，就抛出一个异常
        raise ValueError(f'No assistant with id {id} found in YAML file.')
    
    @classmethod
    def delete_by_id(cls, id: str):

        # 使用绝对路径打开 assistants.yaml 文件
        with open(YamlPathConfig.assistants_yaml_path, 'r') as file:
            data = yaml.safe_load(file) or []

        # 查找具有相同 id 的 assistant
        for i, d in enumerate(data):
            if d['id'] == id:
                # 如果找到了，就删除它
                del data[i]
                break
        else:
            # 如果没有找到，就抛出一个异常
            raise ValueError(f'No assistant with id {id} found in YAML file.')

        # 写回 YAML 文件
        with open(YamlPathConfig.assistants_yaml_path, 'w') as file:
            yaml.dump(data, file)


SDE_PROMPT = """
You are a software engineer working on a project. You are working on a feature implementation task.

"""

REPO_STRUCTURE_PROMPT = """
The project has a file tree structure as follows:

```text
{file_tree}
```

"""

FEATURE_PROMPT = """
Now you are assigned to implement a feature, which is described as follows:

{feature_description}

"""

FOCUS_FILE_PATH_PROMPT = """
Now please provide a list of file paths that you think are relevant to this feature implementation task. 
If there are more than 5 relevant files, you should only provide top 5 files.

Your output should be a title (`##files`, attention: lower case) and a python code block, which include a list of file paths (you are allow to create new files). The example output is as follows:

{format_example}

---

YOU SHOULD ONLY OUTPUT EXISTING FILE PATHS IN THE REPO.
DO NOT INCLUDE ANY OTHER INFORMATION IN YOUR OUTPUT.
YOU SHOULD ONLY OUTPUT TOP 5 FILES.
"""


FOCUS_FILE_PATH_EXAMPLE = """
## files
```python
[
    "path/to/file1",
    "path/to/file2", 
    "path/to/file3",
]
```
"""

FOCUS_FILE_PROMPT = """
Focus files is a list of files that you should focus on for this feature implementation task. 
You should modify these files (or add new file) to implement the feature.
The list of focus files is as follows:

{focus_files}
"""

PLAN_PROMPT = """
Please design a high level plan for this feature implementation task.
Your output should be a title (`## plan`, attention: lower case) and python list of modification plans. The example output is as follows:

{format_example}

---

IF YOU THINK YOU DON'T NEED TO MODIFY A SPECIFIC FILE, YOU SHOULDN'T INCLUDE IT IN YOUR OUTPUT.
DO NOT INCLUDE ANY OTHER INFORMATION IN YOUR OUTPUT.
"""

PLAN_FORMAT_EXAMPLE = """
## plan
```python
[
    ("add", "path/to/file1", "the reason why you want to add this file"),
    ("remove","path/to/file2", "the reason why you want to remove this file"),
    ("modify", "path/to/file3", "the detailed plan for modification"),
    ("modify","path/to/file4", "the detailed plan for modification"),
]
"""

ADD_FILE_PROMPT = """
The path of the file to be added is {file_path}.

The detailed description of the file to be added is as follows:

{action_description}

Please write the file content for me. The output should only be a python string of file content.
DO NOT INCLUDE ANY OTHER INFORMATION EXCEPT THE FILE CONTENT IN YOUR OUTPUT.
"""

MODIFY_FILE_PROMPT = """
The path of the file to be modified is {file_path}, and the content of the file is as follows:

{file_content}

The detailed description of the file to be modified is as follows:

{action_description}

Please write the file content for me. The output should only contain file content.

---

DO NOT CONTAIN QUOTATION MARKS (\", \"\"\", \', etc.) AT THE BEGINNING AND THE END OF THE FILE CONTENT.
YOUR OUTPUT SHOULD JUST USE `\n` FOR LINE BREAK, DO NOT USE `\\n`.
DO NOT INCLUDE ANY OTHER INFORMATION EXCEPT FOR THE FILE CONTENT IN YOUR OUTPUT.
"""



from fastapi import UploadFile, File, Form
from typing import Optional

import pandas as pd

from ..base_node import BaseNode, NodeConfig
from .data_analysis_model import (
    LoadDataInput,
    CleanDataInput,
)
from ...common_models import (
    UserProperties,
    DEFAULT_DATA_ANALYSIS_FOLDER,
)

data_analysis_node_config = {
    "name": "data_analysis",
    "description": "A node for data analysis tasks.",
    "functions": {
        "load_data_from_file": "Load data from a given file.",
        "load_data": "Load data from a given source.",
        "clean_data": "Clean the loaded data.",
    },
}


class DataAnalysisNode(BaseNode):
    config: NodeConfig = NodeConfig(**data_analysis_node_config)

    data: dict[str, pd.DataFrame | str] = None  # To hold the loaded data
    user_properties = None  # To hold the user properties

    def __init__(self):
        super().__init__()
        self.data = {}

    def load_data_from_file(
        self,
        input: UploadFile = File(...),
        properties: UserProperties = Form(...),
    ):
        source_path = (
            DEFAULT_DATA_ANALYSIS_FOLDER
            / properties.user_id
            / properties.session_id
            / input.filename
        )
        source_path.parent.mkdir(parents=True, exist_ok=True)
        self.user_properties = properties
        with open(source_path, "wb") as buffer:
            buffer.write(input.file.read())
        load_data_input = LoadDataInput(
            source_type=input.filename.split(".")[-1].lower(),
            source_path=str(source_path),
        )
        return self.load_data(
            load_data_input=load_data_input,
        )

    def load_data(self, input: LoadDataInput):
        if input.name is None:
            input.name = input.source_path.split("/")[-1].split(".")[0]
        try:
            if input.source_type == "csv":
                self.data[input.name] = pd.read_csv(input.source_path)
            elif input.source_type == "json":
                self.data[input.name] = pd.read_json(input.source_path)
            else:
                self.data[input.name] = input.source_path
            # Add more source types as needed
            return f"Data loaded successfully."
        except Exception as e:
            return f"An error occurred while loading the data: {e}"

    def clean_data(self, input: CleanDataInput):
        try:
            if input.name in self.data.keys():
                return "No data loaded to clean."

            if input.drop_columns:
                self.data[input.name].drop(columns=input.drop_columns, inplace=True)

            if input.fill_na:
                self.data[input.name].fillna(value=input.fill_na, inplace=True)

            return "Data cleaned successfully."
        except Exception as e:
            return f"An error occurred while cleaning the data: {e}"



from abc import ABC, abstractmethod
from enum import Enum
from pydantic import BaseModel, Field

import os
import yaml

class State(str, Enum):
    IDLE = "idle"
    RUNNING = "running"
    DONE = "done"

class BaseToolEntity(ABC):
    @abstractmethod
    def current_state(self):
        pass
    
    @abstractmethod
    def call(self, **kwargs):
        pass

    @abstractmethod
    def need_llm_generate_parameters(self) -> bool:
        pass

    @abstractmethod
    def need_llm_generate_response(self) -> bool:
        pass

    @abstractmethod
    def is_stateful(self) -> bool:
        pass

class FunctionToolEntity(BaseToolEntity):
    parameters: dict[str, any]
    func: callable

    def __init__(self, func: callable):
        self.func = func
        
        self.state = State.IDLE
        self.parameters = {}

    def current_state(self):
        return self.state
    
    def need_llm_generate_parameters(self) -> bool:
        return True

    def need_llm_generate_response(self) -> bool:
        return True

    def is_stateful(self) -> bool:
        return False

    def call(self, **kwargs):
        if self.state == State.IDLE:
            self.state = State.RUNNING
            res = self.func(**kwargs)
            self.state = State.DONE
            return res
        else:
            raise Exception(f"FunctionTool is in state {self.state}, not {State.IDLE}")








from ...utils.singleton import Singleton
from ..common_models import UserProperties, RedisKeyType
import redis
from pathlib import Path
import json
from enum import Enum


@Singleton
class Redis:
    """
    A class for interacting with Redis database.
    """

    def __init__(self):
        """
        Initialize a Redis client instance.
        """
        self.redis = redis.Redis(host="redis", port=6379, decode_responses=True)

    def get(self, key: str) -> str:
        """
        Get the value of a key from Redis.

        Args:
            key (str): The key to retrieve the value for.

        Returns:
            str: The value of the key.
        """
        return self.redis.get(key)

    def set(self, key: str, value: str, save: bool = True) -> str:
        """
        Set the value of a key in Redis.

        Args:
            key (str): The key to set the value for.
            value (str): The value to set for the key.
            save (bool, optional): Whether to save the changes to disk, by default True.

        Returns:
            Optional[str]: If `save` is True, returns the Redis response to the SAVE command.
        """
        self.redis.set(key, value)
        if save:
            return self.redis.save()

    def exists(self, key: str) -> bool:
        """
        Check if a key exists in Redis.

        Args:
            key (str): The key to check for existence.

        Returns:
            bool: True if the key exists, False otherwise.
        """
        return self.redis.exists(key)

    def save(self) -> str:
        """
        Export all key-value pairs to an RDB file.

        Returns:
            Optional[str]: The Redis response to the SAVE command.
        """
        return self.redis.save()

    def safe_get(self, key: str) -> str:
        """
        Get the value of a key from Redis and json loads the value.

        Args:
            key (str): The key to retrieve the value for.

        Returns:
            str: The value of the key.
        """
        return json.loads(self.redis.get(key))

    def safe_set(self, key: str, value, save: bool = True):
        """
        Set the value of a key in Redis and dumps into json before setting.

        Args:
            key (str): The key to set the value for.
            value (str): The value to set for the key.
            save (bool, optional): Whether to save the changes to disk, by default True.

        Returns:
            Optional[str]: If `save` is True, returns the Redis response to the SAVE command.
        """
        self.redis.set(key, json.dumps(value))
        if save:
            return self.redis.save()

    def print_all(self):
        """
        Print all keys and their corresponding values in Redis.
        """
        for key in self.redis.keys():
            print(f"{key}: {self.redis.get(key)}")

    def exists_with_key_type(self, user_properties: UserProperties, type: RedisKeyType):
        """
        Check if a key exists in Redis with a specific key type.

        Args:
            user_properties (UserProperties): The user properties.
            type (RedisKeyType): The Redis key type.

        Returns:
            bool: True if the key exists, False otherwise.
        """
        return self.redis.exists(user_properties.generate_redis_key_with_type(type))

    def safe_get_with_key_type(
        self, user_properties: UserProperties, type: RedisKeyType
    ):
        """
        Get the value of a key from Redis with a specific key type and json loads the value.

        Args:
            user_properties (UserProperties): The user properties.
            type (RedisKeyType): The Redis key type.

        Returns:
            str: The value of the key.
        """
        return json.loads(
            self.redis.get(user_properties.generate_redis_key_with_type(type))
        )

    def safe_set_with_key_type(
        self,
        user_properties: UserProperties,
        type: RedisKeyType,
        value,
        save: bool = True,
    ):
        """
        Set the value of a key in Redis with a specific key type and dumps into json before setting.

        Args:
            user_properties (UserProperties): The user properties.
            type (RedisKeyType): The Redis key type.
            value (str): The value to set for the key.
            save (bool, optional): Whether to save the changes to disk, by default True.

        Returns:
            Optional[str]: If `save` is True, returns the Redis response to the SAVE command.
        """
        self.redis.set(
            user_properties.generate_redis_key_with_type(type), json.dumps(value)
        )
        if save:
            return self.redis.save()



from ..base_node import BaseNode, NodeConfig
from .github_node import GithubNode
from .github_model import (
    RepositoryInput,
    CreatePullRequestInput,
    GetPullRequestByNumberInput,
    GetPullRequestsByQueryInput,
    ModifyPRCommentInput,
)

github_pullrequest_node_config = {
    "name": "github_pullrequest",
    "description": "A node for interacting with GitHub pull requests.",
    "functions": {
        "create_new_pull_request": "Create a new pull request.",
        "get_pull_request_by_number": "Get pull request by number.",
        "get_pull_requests_by_query": "Get pull requests by query.",
        "modify_pr_comment": "Add and modify pull request comment."
    },
}

class GithubPullRequestNode(GithubNode):
    config: NodeConfig = NodeConfig(**github_pullrequest_node_config)

    def __init__(self):
        super().__init__()

    def create_new_pull_request(self, input: CreatePullRequestInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            pr = repo.create_pull(
                title=input.title,
                body=input.body,
                head=input.head,
                base=input.base
            )
            return pr.number
        except Exception as e:
            return str(e)

    def get_pull_request_by_number(self, input: GetPullRequestByNumberInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            pr = repo.get_pull(input.pr_number)
            return {
                "title": pr.title,
                "body": pr.body,
                "state": pr.state,
                "merged": pr.merged
            }
        except Exception as e:
            return str(e)

    def get_pull_requests_by_query(self, input: GetPullRequestsByQueryInput):
        try:
            search_result = self.g.search_issues(input.query, type="pr")
            prs = [{"number": pr.number, "title": pr.title} for pr in search_result]
            return prs
        except Exception as e:
            return str(e)

    def modify_pr_comment(self, input: ModifyPRCommentInput):
        try:
            repo = self.g.get_repo(f"{input.owner}/{input.repo_name}")
            pr = repo.get_pull(input.pr_number)
            comment = pr.get_comment(input.comment_id)
            comment.edit(body=input.body)
            return "Comment modified successfully."
        except Exception as e:
            return str(e)



TOOLS_CHOOSE_PROMPT = """
As a tool selector, you'll provide users with suggestions on tool selection. Depending on the provided tool summary (tools_summary) and user input (input_text), you'll need to follow these steps:

1. Read and understand the tool summary (tools_summary):
   - Understand the features, suitcases, and limitations of each tool.

2. Analyze User Input (input_text):
   - Understand the user's needs or problems.
   - Identify keywords or phrases to determine which tool best suits the user's needs.

3. Decision-making logic:
   - Recommend a tool if the user's needs correspond to the tool's functionality.
   - If the user's needs are not suitable for any tool, or if the information is not sufficient to make a judgment, no tool is recommended.

4. Output:
   - If a tool is recommended, output the tool name (toolname).
   - If no tool is recommended, the output is empty.

Note that recommendations for tool selection should be based on the user's needs and refer to the tool summary provided. Follow the steps above and make sure to provide accurate tool selection suggestions in the output.
"""

TOOLS_CHOOSE_EXAMPLE_PROMPT = """
Here is some examples about tools choosing:

Input:
tools_summary: {
  "ToolA": "For text analysis and data extraction",
  "ToolB": "For image processing and editing",
  ToolC: For audio editing and processing
}
input_text: "I need to analyze a piece of text to extract key information."

Dispose:
- Analyze the input_text and identify the requirements as "text analytics and data extraction".
- Depending on the tools_summary, ToolA matches this requirement.

Output:
{
    "thoughts": {
        "text": "用户需要分析一段文字并提取关键信息。",
        "reasoning": "根据工具概要，ToolA专用于文本分析和数据提取，这与用户的需求相符。",
        "criticism": "虽然用户的需求明确，但仍需考虑文本的具体类型和分析的深度来确保ToolA能够满足需求。",
        "speak": "您希望提取的关键信息是什么类型的？这有助于更好地使用ToolA。"
    },
    "tool": {
        "name": ["ToolA"]
    }
}


Input:
tools_summary: {
  "ToolA": "For text analysis and data extraction",
  "ToolB": "For image processing and editing",
  "ToolC": "For text editing and processing"
}
input_text: "I need to analyze the video and tell me how long it is."

Output:
{
    "thoughts": {
        "text": "用户需要分析视频并了解视频时长。",
        "reasoning": "用户的需求是与视频分析相关的，而提供的工具中没有直接针对视频处理或分析的工具。ToolA和ToolC主要用于文本处理，ToolB则用于图像处理和编辑。",
        "criticism": "虽然ToolB与图像处理有关，但它并不适用于视频分析或处理视频时长这类需求。此外，没有工具被描述为直接支持视频处理功能。",
        "speak": "看来我们目前提供的工具都不适合您的需求，因为它们主要是用于文本和图像处理，而不是视频分析。"
    },
    "tool": {
        "name": [
            
        ]
    }
}

Input:
tools_summary: {
"ToolA": "For text analysis and data extraction",
"ToolB": "For image processing and editing",
"ToolC": "For creating interactive web applications"
}
input_text: "I need to analyze survey data and create an interactive web dashboard to display the results."

Output:
{
    "thoughts": {
        "text": "用户需要分析调查数据并创建一个交互式的网络仪表板来展示结果。",
        "reasoning": "根据工具摘要，ToolA 适用于文本分析和数据提取，而 ToolC 适用于创建交互式的网络应用程序。",
        "criticism": "虽然 ToolA 适用于数据分析，但无法创建网络仪表板；ToolC 可以创建交互式的网络应用，但可能不具备分析数据的功能。",
        "speak": "根据您的需求，ToolA 和 ToolC 都可能是有用的工具，但需要结合使用才能满足您的需求。"
    },
    "tool": {
        "name": ["ToolA", "ToolC"]
    }
}



"""

TOOLS_CHOOSE_HINT = """
You should only respond in JSON format as described below 
Response Format: 

{{{{
    "thoughts": {{{{
        "text": "your thoughts in the current context",
        "reasoning": "reasoning for tool selection and input content",
        "criticism": "critical thinking on tool selection and input in current context",
        "speak": "words you want to speak to the user",
    }}}},
    "tool": {{{{
        "name": ['tool_name'], 
    }}}}
}}}}

The strings corresponding to "text", "reasoning", "criticism", and "speak" in JSON should be described in Chinese.

If you don't need to use a tool(like solely chat scene), or have already reasoned the final answer associated with user input from the tool, You must abide by the following rules: 
1. The tool's name in json is [].

Do not output any other information and do not contain quotation marks, such as `, \", \' and so on.
Ensure the output can be parsed by Python json.loads.
Don't output in markdown format, something like ```json or ```,just output in the corresponding string format

"""




import logging


def get_logger(logger_name, log_level=logging.INFO):
    logger = logging.getLogger(logger_name)
    logger.setLevel(log_level)
    logger.propagate = False  # Prevent log propagation to avoid double logging

    formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(name)s: %(message)s")

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)

    if not logger.handlers:
        logger.addHandler(console_handler)

    return logger



import subprocess
import queue
import threading
import time

from typing import Optional
from pydantic import BaseModel, Field


class PythonProcess(BaseModel):
    # TODO: check arg legal
    working_dir: str = Field(default="/tmp", description="Working directory.")
    args: list[str] = Field(default=[], description="Arguments for python process.")
    python_exec: str = Field(default="python3", description="Python executable path.")
    python_file_path: str = Field(default="main.py", description="Python file path.")

    def run(self):
        cmd = [self.python_exec] + self.args + [self.python_file_path]
        return subprocess.run(
            [
                "bash -c 'cd {dir} && {real_cmd}'".format(
                    dir=self.working_dir, real_cmd=" ".join(cmd)
                ),
            ],
            shell=True,
            stderr=subprocess.PIPE,
        )

    #     cmd = [self.python_exec] + self.args + [self.python_file_path]

    #     proc = subprocess.Popen(
    #         [
    #             "bash -c 'cd {dir} && {real_cmd}'".format(
    #                 dir=self.working_dir, real_cmd=" ".join(cmd)
    #             ),
    #         ],
    #         shell=True,
    #         stdout=subprocess.PIPE,
    #         stderr=subprocess.PIPE,
    #     )

    #     stdout_queue = queue.Queue()
    #     stderr_queue = queue.Queue()

    #     stdout_th = threading.Thread(
    #         target=self._output_reader, args=(1, proc, stdout_queue)
    #     )
    #     stderr_th = threading.Thread(
    #         target=self._output_reader, args=(2, proc, stderr_queue)
    #     )

    #     stdout_th.start()
    #     stderr_th.start()

    #     stdout_buf = ""
    #     stderr_buf = ""

    #     try:
    #         while proc.poll() is None:
    #             try:
    #                 line = stdout_queue.get(block=False)
    #                 print(line, end="")
    #                 stdout_buf += line
    #                 line = stderr_queue.get(block=False)
    #                 print(line, end="")
    #                 stderr_buf += line
    #             except queue.Empty:
    #                 pass
    #             time.sleep(0.002)
    #     finally:
    #         proc.terminate()
    #         try:
    #             proc.wait(timeout=0.5)
    #         except subprocess.TimeoutExpired:
    #             pass

    #     stdout_th.join()
    #     stderr_th.join()

    #     return subprocess.CompletedProcess(
    #         args=proc.args,
    #         returncode=proc.returncode,
    #         stdout=stdout_buf,
    #         stderr=stderr_buf,
    #     )

    # def _output_reader(self, id, proc, out_queue):
    #     if id == 1:
    #         for line in iter(proc.stdout.readline, b""):
    #             out_queue.put(line.decode("utf-8"))
    #     elif id == 2:
    #         for line in iter(proc.stderr.readline, b""):
    #             out_queue.put(line.decode("utf-8"))



from pydantic import BaseModel

# Base input for most repository-related functions
class RepositoryInput(BaseModel):
    owner: str
    repo_name: str

# Specific file content
class GetSpecificContentFileInput(RepositoryInput):
    path: str

# Create a new file
class CreateFileInput(RepositoryInput):
    path: str
    message: str
    content: str  # Content should be base64 encoded.

# Edit a file
class EditFileInput(CreateFileInput):
    sha: str

# Delete a file
class DeleteFileInput(RepositoryInput):
    path: str
    message: str
    sha: str

# Base input for getting an issue
class GetIssueInput(BaseModel):
    owner: str
    repo_name: str
    issue_number: int

# Create comment on an issue
class CreateIssueCommentInput(BaseModel):
    owner: str
    repo_name: str
    issue_number: int
    comment_content: str

# Base input for creating an issue
class CreateIssueInput(BaseModel):
    owner: str
    repo_name: str
    title: str

# Create an issue with body
class CreateIssueWithBodyInput(CreateIssueInput):
    body: str

# Create an issue with labels
class CreateIssueWithLabelsInput(CreateIssueInput):
    labels: list[str]

# Create an issue with assignee
class CreateIssueWithAssigneeInput(CreateIssueInput):
    assignee: str

# Create an issue with milestone
class CreateIssueWithMilestoneInput(CreateIssueInput):
    milestone_number: int  # Milestone number

class CloseAllIssuesInput(BaseModel):
    owner: str
    repo_name: str

# Input for creating a commit status check
class CreateCommitStatusCheckInput(BaseModel):
    owner: str
    repo_name: str
    sha: str
    state: str  # error, failure, pending, or success
    target_url: str
    description: str
    context: str

# Input for getting the commit date
class GetCommitDateInput(BaseModel):
    owner: str
    repo_name: str
    sha: str

# Base input for most branch-related functions
class RepositoryInput(BaseModel):
    owner: str
    repo_name: str

# Input for getting a specific branch
class GetBranchInput(RepositoryInput):
    branch_name: str

# Input for getting the HEAD commit of a branch
class GetBranchHeadCommitInput(GetBranchInput):
    pass

# Input for getting the protection status of a branch
class GetBranchProtectionStatusInput(GetBranchInput):
    pass

# Input for seeing the required status checks of a branch
class GetBranchRequiredStatusChecksInput(GetBranchInput):
    pass   


# Base input for most pull request-related functions
class RepositoryInput(BaseModel):
    owner: str
    repo_name: str

# Input for creating a new pull request
class CreatePullRequestInput(RepositoryInput):
    title: str
    body: str
    head: str  # The name of the branch where your changes are implemented
    base: str  # The name of the branch you want the changes pulled into

# Input for getting a pull request by number
class GetPullRequestByNumberInput(RepositoryInput):
    pr_number: int

# Input for getting pull requests by query
class GetPullRequestsByQueryInput(RepositoryInput):
    query: str

# Input for adding and modifying a pull request comment
class ModifyPRCommentInput(RepositoryInput):
    pr_number: int
    comment_id: int
    body: str

# Base input for most milestone-related functions
class RepositoryInput(BaseModel):
    owner: str
    repo_name: str

# Input for getting a specific milestone
class GetMilestoneInput(RepositoryInput):
    milestone_number: int

# Input for creating a milestone
class CreateMilestoneInput(RepositoryInput):
    title: str

# Input for creating a milestone with state and description
class CreateMilestoneWithDetailsInput(CreateMilestoneInput):
    state: str  # Either "open", "closed", or "all"
    description: str

# Base input for getting repositories of a user
class GetUserRepositoriesInput(BaseModel):
    username: str

# Input for inviting a user to an organization
class InviteUserToOrgInput(BaseModel):
    org_name: str
    username: str

# Base input for all search operations
class BaseSearchInput(BaseModel):
    query: str

# Extended search models for specific search types
class SearchCodeInput(BaseSearchInput):
    sort: str = None
    order: str = None

class SearchCommitsInput(BaseSearchInput):
    sort: str = None
    order: str = None

class SearchIssuesAndPRsInput(BaseSearchInput):
    sort: str = None
    order: str = None

class SearchLabelsInput(BaseSearchInput):
    repository_id: int
    sort: str = None
    order: str = None

class SearchRepositoriesInput(BaseSearchInput):
    sort: str = None
    order: str = None

class SearchTopicsInput(BaseSearchInput):
    sort: str = None
    order: str = None

class SearchUsersInput(BaseSearchInput):
    sort: str = None
    order: str = None


from pydantic import BaseModel
from typing import List, Dict, Union, Optional


class LoadDataInput(BaseModel):
    name: Optional[str]  # Name of the file, if no provided, use the file name
    source_type: str  # 'csv', 'json', etc.
    source_path: str  # File path or URL


class CleanDataInput(BaseModel):
    name: str  # Name of the file to clean
    drop_columns: Optional[List[str]]  # List of column names to drop
    fill_na: Optional[
        Dict[str, Union[str, float, int]]
    ]  # Dictionary specifying what value to replace NaN with for each column



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['yorg_test',
 'yorg_test.core',
 'yorg_test.core.agents',
 'yorg_test.core.agents.data_scientist',
 'yorg_test.core.agents.software_engineer',
 'yorg_test.core.assistant',
 'yorg_test.core.assistant.prompt',
 'yorg_test.core.assistant.tools',
 'yorg_test.core.nodes',
 'yorg_test.core.nodes.code_runner',
 'yorg_test.core.nodes.data_analysis',
 'yorg_test.core.nodes.document_loader',
 'yorg_test.core.nodes.git',
 'yorg_test.core.nodes.github',
 'yorg_test.core.nodes.openai',
 'yorg_test.core.service',
 'yorg_test.utils',
 'yorg_test.utils.code_executor']

package_data = \
{'': ['*']}

install_requires = \
['include']

setup_kwargs = {
    'name': 'phireact',
    'version': '0.0.1',
    'description': 'test package for yorg',
    'long_description': '# YORG Test Package\n\nthis is a test package\n\n# test version\n\n## how to use\n\n- Set up test version package\n``` shell\npip install -i https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple yorg-test\n```\n- Set up openapikey\n(If you are in China, please set up a proxy to ensure that you can connect to openai.)\n```python\nimport os\n#os.environ[\'http_proxy\'] = \'http://127.0.0.1:10809\'  # 这里设置自己的代理端口号\n#os.environ[\'https_proxy\'] = \'http://127.0.0.1:10809\'  # 这里设置自己的代理端口号\nos.environ[\'OPENAI_CHAT_API_KEY\'] = \'sk-br3j7Gxxxxxxxxvt8r\'\n```\n- set up yaml file\nWe have some tools built in eg.code_interpreter,swe_tool\n\ntools.yaml\n```yaml\nYORG: 0.0.1\ninfo:\n  title: yorg_tools_document\n  description: yorg tool define document.\n  version: \'v1\'\ntools:\n  code_interpreter:\n    name: code_interpreter\n    entity_name: code_interpreter\n    summary: Run the code through code_interpreter and return the result of the code run. If your query is about math, computer science, or data science, you can use this tool to get the result of the code run.\n    parameters:\n      - name: code\n        description: code text that requires code_interpreter to run\n        required: true\n        parameter_schema:\n          type: string\n    responses:\n      success:\n        description: OK\n        content:\n          result:\n            type: string\n            description: the result of the code run\n  example_stateful_tool:\n    name: example_stateful_tool\n    entity_name: example_stateful_tool\n    summary: This tool is an example of a stateful tool. It will get two number from user input and return the sum of the two numbers.\n    parameters: []\n    responses: {}  \n  swe_tool:\n    name: sew_tool\n    entity_name: swe_tool\n    summary: SoftWare Engineer Agent(swe_tool) specializes in working with code files.\n    parameters: []\n    responses: {}  \n```\nTools are categorized into stateful tools and function tools,\nFunction tools can describe parameters and return values directly in tools.yaml\nFunctions can be registered using decorators\n```python\nfrom yorg_test.core.assistant.tools.tools import register_function_tool\n@register_function_tool\ndef code_test(code: str):\n    return {\n        "type": "success",\n        "content": {\n            "result": "Hello, World!",\n        },\n    }\n```\nIf it\'s a stateful tool you need to write an additional yaml file with a stateful description\n```yaml\nstart_stage: "init"\nfinish_stage: "finish"\nall_stages:\n  init:\n    name: "init"\n    next_stage_entry: \n      stage_1:\n        - name: x\n          required: true\n          parameter_schema:\n            type: number\n            description: "input value x"\n    need_llm_generate_parameters: false\n    need_llm_generate_response: false\n  stage_1:\n    name: "stage_1"\n    next_stage_entry: \n      stage_2:\n        - name: y \n          required: true\n          parameter_schema:\n            type: number\n            description: "input value y"\n    need_llm_generate_parameters: false \n    need_llm_generate_response: false\n  stage_2:\n    name: "stage_2"\n    next_stage_entry: \n      stage_3: []\n    need_llm_generate_parameters: false\n    need_llm_generate_response: false\n  stage_3:\n    name: "stage_3"\n    next_stage_entry: \n      finish: []\n    need_llm_generate_parameters: false\n    need_llm_generate_response: false\n  finish:\n    name: "finish"\n    next_stage_entry: {}\n    need_llm_generate_parameters: false\n    need_llm_generate_response: false\n```\nStateful tools can also be registered using decorators.\nThe yaml file is registered in init.\n```python\nfrom yorg_test.core.assistant.tools.tools import register_stateful_tool\nfrom yorg_test.core.assistant.tools.stateful_tool_entity import StatefulToolEntity\n@register_stateful_tool\nclass ExampleStatefulToolEntity(StatefulToolEntity):\n    """\n    This example tool entity is stateful, and it has 3 inner stages.\n\n    stage1: take integer x as input\n    stage2: take integer y as input\n    stage3: no input, return x + y\n    """\n    def __init__(self):\n        super().__init__("example_stateful_tool.yaml")\n    def _call(self, **kwargs):\n        if "goto" not in kwargs:\n            if self.current_stage.name == self.config.start_stage:\n                return {\n                    "type": "success",\n                    "content": {"message": "stateful tool is started"},\n                }\n            else:\n                return {\n                    "type": "error",\n                    "content": {"message": "please provide `goto` parameter"},\n                }\n        request_next_stage = kwargs["goto"]\n        if request_next_stage not in self.config.all_stages:\n            return {\n                "type": "error",\n                "content": {"message": f"stage {request_next_stage} not found"},\n            }\n        self.current_stage = self.config.all_stages[request_next_stage]\n        match self.current_stage.name:\n            case "stage_1":\n                return self._stage1(kwargs["x"])\n            case "stage_2":\n                return self._stage2(kwargs["y"])\n            case "stage_3":\n                return self._stage3()\n            case self.config.finish_stage:\n                return self._finish()\n            case _:\n                return {\n                    "type": "error",\n                    "content": {\n                        "message": f"stage {self.current_stage.name} not found"\n                    },\n                }\n    def _stage1(self, x: int):\n        self.x = x\n        return {"type": "success", "content": {"message": "stage1 done"}}\n    def _stage2(self, y: int):\n        self.y = y\n        return {"type": "success", "content": {"message": "stage2 done"}}\n    def _stage3(self):\n        return {"type": "success", "content": {"result": self.x + self.y}}\n    def _finish(self):\n        return {"type": "success", "content": {"message": "stateful tool is finished"}}\n```\n- run example\n\n```python\nimport yorg_test\nthreads = yorg_test.Threads.create(\'tools.yaml\')\nassistant = yorg_test.Assistants.create(name="Test Assistant", model="gpt-4-1106-preview", instructions="Use swe tool auto fix code files", tools=[{\'type\':\'swe_tool\'}])\nprint(assistant.id)\n# 运行 Threads 对象\nresult = threads.run(assistant.id, "Use SoftWare Engineer Agent swe tool auto fix code files.")\nprint(result)\n\nresult = threads.run(assistant.id, "the repo url is https://github.com/YORG-AI/Open-Assistant",goto="stage_1")\nprint(result)\n\nresult = threads.run(assistant.id, "add helloworld feature to readme",  goto="stage_2")\nprint(result)\n\nresult = threads.run(assistant.id, "focus_files_name_list = [README.md]", goto="stage_3")\nprint(result)\n\nresult = threads.run(assistant.id, "action=3", goto="stage_4")\nprint(result)\n\nresult = threads.run(assistant.id, "", goto="stage_5")\nprint(result)\n\nresult = threads.run(assistant.id, "action=0,action_idx=0", goto="stage_6")\nprint(result)\n\nresult = threads.run(assistant.id, "", goto="finish")\nprint(result)\n```\n\n',
    'author': 'zxy',
    'author_email': 'zhongxingyuemail@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
